#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "align.h"

#define MAX_STR_LEN 1200

#ifndef DEFAULT_SCORE_FILE
#define DEFAULT_SCORE_FILE "./dat/PAM-250.score"
#endif

void version()
{
  printf("   Flexible Multiple Alignment Program - Version 0.34 alpha\n");
  printf("         - Suboptimal and Parametric Analysis -\n");
}

void main(int argc, char *argv[])
{
  FILE *score_fp, *score_fp2, *in_fp, *w1_fp, *w2_fp;
  char buf[MAX_STR_LEN], *score_file, *score_file2, *w1_file, *w2_file;
  int kbs, k, fs_k, bs_k;
  int dir_mask;
  int delta, lbound, lbound1, lbound2, accel;
  int dlt;
  int ngroup, no_ope, line;
  int i, j, n, m, p, q, r, nseq, use_nseq, dim;
  int score, min_score, min_score2, min_output;
  int score_question;
  int use_astar, use_pairopt, use_ub, use_inc;
  int use_step, use_accel, use_group, use_delta;
  int use_k, use_skip, use_detour, use_weight;
  int use_parametric_gap, use_parametric_weight;
  int use_parametric_score, use_ub_parametric;
  int auto_gap_penalty, make_positive;
  int print_subopt;
  int *used_sequence;
  int use_change_score;
  char tmp_char, tmp_char1, tmp_char2;
  char changed_pair1, changed_pair2;
  int changed_score;

  int start_gap_penalty, start_gap_denominator;
  int end_gap_penalty, end_gap_denominator;
  solutions_p boundaries;
  solutions2_p boundaries2, boundaries3;
  int next_iteration, max_iteration;

  sequence_t *seq, *tmp_seq;
  alignment_t al[MAX_DIM], *result_al_p;
#ifdef VERBOSE
  time_t t_start, t_end;
#endif

  version();

  seq = (sequence_t *) malloc (MAX_NSEQ * sizeof(sequence_t));
  in_fp = NULL;
  kbs = 0;
  dlt = 100;
  delta = 0;
  lbound = 0;
  lbound1 = 0;
  lbound2 = 0;
  k = 1;
  fs_k = 1;
  bs_k = 0;
  dir_mask = FMASK;
  use_nseq = MAX_NSEQ;
  use_delta = 0;
  use_k = 0;
  use_astar = 1;
  use_pairopt = 1;
  use_step = 0;
  use_skip = 0;
  use_detour = 0;
  used_sequence = NULL;
  use_ub = 0;
  use_inc = 0;
  use_accel = 0;
  use_group = 0;
  min_output = 0;
  auto_gap_penalty = 1;
  gap_denominator = 1;
  make_positive = 0;
  print_subopt = 0;
  use_parametric_gap = 0;
  use_weight = 0;
  use_parametric_weight = 0;
  use_parametric_score = 0;
  use_ub_parametric = 0;
  use_change_score = 0;
  score_question = 0;
  boundaries = NULL;
  boundaries2 = NULL;
  boundaries3 = NULL;
  max_iteration = 0;
  next_iteration = 1;
  no_ope = 0;
  line=50;
  score_file = DEFAULT_SCORE_FILE;

  init_scores();

  while((--argc > 0) && (**++argv == '-')){
    switch(*++*argv){
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      use_nseq = atoi(*argv);
      if(use_nseq > MAX_NSEQ){
	use_nseq = MAX_NSEQ;
      }
      if ((argc <= 1) || (**(argv+1) == '-')) break;
      use_skip = 1; m = 0; p = 0;
      used_sequence = (int *)malloc((use_nseq+1)*sizeof(int));
      while ((argc > 1) && (**(argv+1) != '-')) {
	argc--; argv++;
	used_sequence[m] = atoi(*argv) - 1;
	if (p < used_sequence[m]) p = used_sequence[m];
	m++;
	if (m > use_nseq) {
	  fprintf(stderr,"Error: Too many numbers of sequences\n");
	  exit(-1);
	}
      }
      if (m < use_nseq) {
	fprintf(stderr,"Error: Too few numbers of sequences\n");
	exit(-1);
      }
      m = 0;
      break;
    case 'a':
      use_astar = 1;
      break;
    case 'A':
      use_astar = 0;
    case 'b':
      dir_mask = FMASK | BMASK;
      break;
    case 'c':
      switch((*argv)[1]){
      case '\0':
	score_file = *++argv;
	argc--;
	break;
      case 'c':   /* must use without -g option */
	use_parametric_score = 1;
	score_file = *++argv;
	score_file2 = *++argv;
	argc -= 2;
	if ((boundaries3 = (solutions2_p)malloc(MAXSOLUTIONS * sizeof(solutions2_t))) == NULL) {
	  fprintf(stderr,"Can't allocate memory for solutions of parametric problem\n");
	  exit(1);
	}
	max_iteration = MAXSOLUTIONS;
	next_iteration = 2;
	break;
      case 'q':   /* please ignore */
	score_question = 1;
	break;
      }
      break;
    case 'C':  /* Do not use with -cc option */
      use_change_score = 1;
      changed_pair1 = (*++argv)[0];
      changed_pair2 = (*++argv)[0];
      changed_score = atoi(*++argv);
      argc -= 3;
      break;
    case 'd':   /*  detour-like suboptimal solutions  */
      use_detour = 1;
      break;
    case 'D':
      use_delta = 1;
      delta = atoi(*++argv);
      argc--;
      break;
    case 'e':
      use_accel = 1;
      accel = atoi(*++argv);
      argc--;
      break;
    case 'f':
      if(strcmp(*++argv, "-") == 0){
	in_fp = stdin;
	if(!min_output){
	  printf("Input file: standard input\n");
	}
      }else{
	if((in_fp = fopen(*argv, "rt")) == NULL){
	  fprintf(stderr, "Can't open %s\n", *argv);
	  exit(1);
	}
	if(!min_output){
	  printf("Input file: %s\n", *argv);
	}
      }
      argc--;
      break;
    case 'g':
      switch((*argv)[1]){
      case '\0':
	internal_gap_penalty = atoi(*++argv);
	external_gap_penalty = atoi(*++argv);
	auto_gap_penalty = 0;
	argc -= 2;
	break;
      case 'g':
	start_gap_penalty = atoi(*++argv);
	start_gap_denominator = atoi(*++argv);
	end_gap_penalty = atoi(*++argv);
	end_gap_denominator = atoi(*++argv);
	use_parametric_gap = 1;
	if ((boundaries = (solutions_p)malloc(MAXSOLUTIONS * sizeof(solutions_t))) == NULL) {
	  fprintf(stderr,"Can't allocate memory for solutions of parametric problem\n");
	  exit(1);
	}
	max_iteration = MAXSOLUTIONS;
	next_iteration = 2;
	argc -= 4;
	break;
      }
      break;
    case 'G':
      gap_denominator = atoi(*++argv);
      argc--;
      break;
    case 'h':
      printf("== command line options ==\n");
      printf("  -#number_of_sequences (#, #, ...)\n");
      printf("  -a <use astar: default> -A <not astar>\n");
      printf("  -b <bidirectional>      -c score_file\n");
      printf("  -cc score_file1 score_file2 <parametric>\n");
      printf("  -C char_1 char_2 new_score <change score of a pair>\n");
      printf("  -d <smart suboptimal enumeration>\n");
      printf("  -D (delta * gap_denominator) <suboptimal>\n");
/*      printf(" (-e accell)\n");*/
      printf("  -f input_file\n");
      printf("  -g internal_gap_cost external_gap_cost\n");
      printf("  -gg gap_cost1 gap_denominator1\n");
      printf("      gap_cost2 gap_denominator2 <parametric>\n");
      printf("  -G gap_denominator\n");
      printf("  -h <this help>\n");
/*      printf("  -K #solutions\n");
      printf(" (-i delta <inc>)\n");*/
      printf("  -l characters_on_one_line (default 50)\n");
      printf("  -k wk fs_k bs_k         -m <output min>\n");
      printf("  -n <no operation>\n");
      printf("  -o <pairopt>            -O <not pairopt> \n");
      printf("  -p <make positive>      -q #groups\n");
      printf("  -P <print suboptimal solutions>\n");
      printf("  -r <backward>           -s <step>\n");
      printf("  -u (upper_bound * gap_denominator) \n");
      printf("  -uu upper_bound1 upper_bound2\n");
      printf("  -v <version> -w weight_file\n");
      printf("  -ww weight_file1, weight_file2\n");
      exit(0);
      break;
    case 'i':
      use_inc = 1;
      dlt = atoi(*++argv);
      argc--;
      break;
    case 'k':
      k = atoi(*++argv);
      fs_k = atoi(*++argv);
      bs_k = atoi(*++argv);
      argc -= 3;
      break;
    case 'K':
      use_k=1;
      kbs = atoi(*++argv);
      argc--;
      break;
    case 'l':
      line = atoi(*++argv);
      argc--;
      break;
    case 'm':
      min_output = 1;
      break;
    case 'n':
      no_ope = 1;
      break;
    case 'o':
      use_pairopt = 1;
      break;
    case 'O':
      use_pairopt = 0;
      break;
    case 'p':
      make_positive = 1;
      break;
    case 'P':
      print_subopt = 1;
      break;
    case 'q':
      use_group = 1;
      ngroup = atoi(*++argv);
      argc--;
      break;
    case 'r':
      dir_mask = BMASK;
      break;
    case 's':
      use_step = 1;
      break;
    case 'u':
      switch((*argv)[1]){
      case '\0':
	use_ub = 1;
	lbound = atoi(*++argv);
	argc--;
	break;
      case 'u':
	use_ub_parametric = 1;
	lbound1 = atoi(*++argv);
	lbound2 = atoi(*++argv);
	argc -= 2;
	break;
      }
      break;
    case 'v':
      version();
      exit(0);
      break;
    case 'w':
      switch((*argv)[1]){
      case '\0':
	use_weight = 1;
	w1_file = *++argv;
	argc--;
	break;
      case 'w':
	use_parametric_weight = 1;
	w1_file = *++argv;
	w2_file = *++argv;
	if ((boundaries2 = (solutions2_p)malloc(MAXSOLUTIONS * sizeof(solutions2_t))) == NULL) {
	  fprintf(stderr,"Can't allocate memory for solutions of parametric problem\n");
	  exit(1);
	}
	max_iteration = MAXSOLUTIONS;
	next_iteration = 2;
	argc -= 2;
	break;
      }
      break;
    }
  }
  if((score_fp = fopen(score_file, "rt")) == NULL){
    fprintf(stderr, "Can't open %s\n", score_file);
    exit(-1);
  }
  fgets(buf, 'Z' - 'A', score_fp);
  for(n = 0; buf[n] != '\n'; n++){
    if((buf[n] < 'A') || (buf[n] > 'Z')){
      fprintf(stderr, "Invalid character: %c", buf[n]);
      exit(-1);
    }
  }
  fscanf(score_fp, "%d", &score);
  min_score = max_score = score;
  score_table[buf[0] - 'A'][buf[0] - 'A'] = score;
  if (use_parametric_score) score_table1[buf[0] - 'A'][buf[0] - 'A'] = score;
  for(i = 1; i < n; i++){
    for(j = 0; j < i; j++){
      fscanf(score_fp, "%d", &score);
      if(score < min_score){
	min_score = score;
      }
      if(score > max_score){
	max_score = score;
      }
      score_table[buf[i] - 'A'][buf[j] - 'A'] = score;
      score_table[buf[j] - 'A'][buf[i] - 'A'] = score;
      if (use_parametric_score) {
	score_table1[buf[i] - 'A'][buf[j] - 'A'] = score;
	score_table1[buf[j] - 'A'][buf[i] - 'A'] = score;
      }
    }
    fscanf(score_fp, "%d", &score);
    if(score < min_score){
      min_score = score;
    }
    if(score > max_score){
      max_score = score;
    }
    score_table[buf[i] - 'A'][buf[i] - 'A'] = score;
    if (use_parametric_score) score_table1[buf[i] - 'A'][buf[i] - 'A'] = score;
  }
  fclose(score_fp);
  max_score1 = max_score;

  if (use_parametric_score) {
    if((score_fp2 = fopen(score_file2, "rt")) == NULL){
      fprintf(stderr, "Can't open %s\n", score_file2);
      exit(-1);
    }
    fgets(buf, 'Z' - 'A', score_fp2);
    for(n = 0; buf[n] != '\n'; n++){
      if((buf[n] < 'A') || (buf[n] > 'Z')){
	fprintf(stderr, "Invalid character: %c", buf[n]);
	exit(-1);
      }
    }
    fscanf(score_fp2, "%d", &score);
    max_score2 = min_score2 = score;
    score_table2[buf[0] - 'A'][buf[0] - 'A'] = score;
    for(i = 1; i < n; i++){
      for(j = 0; j < i; j++){
	fscanf(score_fp2, "%d", &score);
	if(score < min_score2){
	  min_score2 = score;
	}
	if(score > max_score2){
	  max_score2 = score;
	}
	score_table2[buf[i] - 'A'][buf[j] - 'A'] = score;
	score_table2[buf[j] - 'A'][buf[i] - 'A'] = score;
      }
      fscanf(score_fp2, "%d", &score);
      if(score > max_score2){
	max_score2 = score;
      }
      if(score > max_score2){
	max_score2 = score;
      }
      score_table2[buf[i] - 'A'][buf[i] - 'A'] = score;
    }
    fclose(score_fp2);
  }

  if (use_change_score) {
    score_table[changed_pair1 - 'A'][changed_pair2 - 'A'] = changed_score;
    if (changed_pair1 != changed_pair2)
      score_table[changed_pair2 - 'A'][changed_pair1 - 'A'] = changed_score;
    fprintf(stderr, " score of %c-%c was changed to %d\n",
	    changed_pair1, changed_pair2, changed_score);
  }

  if(auto_gap_penalty){
    internal_gap_penalty = external_gap_penalty = min_score;
    if (use_parametric_score) {
      internal_gap_penalty1 = external_gap_penalty1 = min_score;
      internal_gap_penalty2 = external_gap_penalty2 = min_score2;
    }
  }/* else{
    if(internal_gap_penalty > max_score){
      max_score = internal_gap_penalty;
    }
    if(external_gap_penalty > max_score){
      max_score = external_gap_penalty;
    }
  }*/           /* deleted --- because gap penalty will vary:  */
                /* added to every subroutine */

  for(i = 'A'; i <= 'Z'; i++){
    score_table[i - 'A'][INTERNAL_GAP_CHAR - 'A'] = internal_gap_penalty;
    score_table[INTERNAL_GAP_CHAR - 'A'][i - 'A'] = internal_gap_penalty;
    if (use_parametric_score) {
      score_table1[i - 'A'][INTERNAL_GAP_CHAR - 'A'] = internal_gap_penalty;
      score_table1[INTERNAL_GAP_CHAR - 'A'][i - 'A'] = internal_gap_penalty;
      score_table2[i - 'A'][INTERNAL_GAP_CHAR - 'A'] = min_score2;
      score_table2[INTERNAL_GAP_CHAR - 'A'][i - 'A'] = min_score2;
    }
  }
  for(i = 'A'; i <= 'Z'; i++){
    score_table[i - 'A'][EXTERNAL_GAP_CHAR - 'A'] = external_gap_penalty;
    score_table[EXTERNAL_GAP_CHAR - 'A'][i - 'A'] = external_gap_penalty;
    if (use_parametric_score) {
      score_table1[i - 'A'][EXTERNAL_GAP_CHAR - 'A'] = external_gap_penalty;
      score_table1[EXTERNAL_GAP_CHAR - 'A'][i - 'A'] = external_gap_penalty;
      score_table2[i - 'A'][EXTERNAL_GAP_CHAR - 'A'] = min_score2;
      score_table2[EXTERNAL_GAP_CHAR - 'A'][i - 'A'] = min_score2;
    }
  }
  score_table[INTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
  score_table[INTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
  score_table[EXTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
  score_table[EXTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
  if (use_parametric_score) {
    score_table1[INTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
    score_table1[INTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
    score_table1[EXTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
    score_table1[EXTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
    score_table2[INTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
    score_table2[INTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
    score_table2[EXTERNAL_GAP_CHAR - 'A'][INTERNAL_GAP_CHAR - 'A'] = 0;
    score_table2[EXTERNAL_GAP_CHAR - 'A'][EXTERNAL_GAP_CHAR - 'A'] = 0;
  }

  if(!make_positive){
    max_score = 0;
    if (use_parametric_score) {
      max_score1 = 0;
      max_score2 = 0;
    }
  }

  if (score_question) {
    while(1) {
      printf("pair > ");
      scanf("%c%c", &tmp_char1, &tmp_char2);
      printf("score = %d\n", score_table[tmp_char1 - 'A'][tmp_char2 - 'A']);
      scanf("%c", &tmp_char1);
    }
  }

  nseq = 0;
  if(in_fp != NULL){
    if (use_skip) {
      while(fgets(buf, MAX_STR_LEN, in_fp) != NULL){
	if((buf[0] == '\n') || (buf[0] == '#')){
	  continue;
	}
	seq[nseq].len = strlen(buf) - 1;
	buf[seq[nseq].len] = '\0';
	if((seq[nseq].str = strdup(buf)) == NULL){
	  fprintf(stderr, "Can't allocate memory for sequence\n");
	  exit(-1);
	}
	if(++nseq > p){
	  break;
	}
      }
    } else {
      while(fgets(buf, MAX_STR_LEN, in_fp) != NULL){
	if((buf[0] == '\n') || (buf[0] == '#')){
	  continue;
	}
	seq[nseq].len = strlen(buf) - 1;
	buf[seq[nseq].len] = '\0';
	if((seq[nseq].str = strdup(buf)) == NULL){
	  fprintf(stderr, "Can't allocate memory for sequence\n");
	  exit(-1);
	}
	if(++nseq > use_nseq){
	  break;
	}
      }
    }
    fclose(in_fp);
  }
  while(argc > 0){
    seq[nseq].len = strlen(*argv);
    seq[nseq].str = *argv;
    if(++nseq > use_nseq){
      break;
    }
    argc--;
    argv++;
  }
  if(nseq == 0){
    exit(0);
  }

  if(use_skip) {
#ifdef VERBOSE
    printf("Used sequences:");
#endif
    if (p >= nseq) {
      fprintf(stderr, "Too large number of sequence\n");
      fprintf(stderr, "Using first %d sequences\n", use_nseq);
      nseq = use_nseq;
    }
    tmp_seq = seq;
    seq = (sequence_t *)malloc(MAX_NSEQ * sizeof(sequence_t));
    
    for (i=0;i<use_nseq;i++) {
      seq[i].len = tmp_seq[used_sequence[i]].len;
      seq[i].str = tmp_seq[used_sequence[i]].str;
#ifdef VERBOSE
      printf(" %d", used_sequence[i]);
#endif
    }
#ifdef VERBOSE
    printf("\n");
#endif
    free(tmp_seq);
  }

  if(nseq > use_nseq){
    fprintf(stderr, "Using only first %d sequences\n", use_nseq);
    nseq = use_nseq;
  }
#ifdef VERBOSE
  t_start = time(NULL);
#endif

  if ((weights = (int **)malloc(nseq*sizeof(int *))) == NULL) {
    fprintf(stderr,"Could not allocate memory for wight matrix\n");
    exit(1);
  }
  for (q = 0;q < nseq;q++) {
    if ((weights[q] = (int *)malloc(nseq*sizeof(int))) == NULL) {
      fprintf(stderr,"Could not allocate memory for wight matrix\n");
      exit(1);
    }
    for (r = 0;r < nseq;r++) {
      if (q==r) weights[q][r] = 0;
      else weights[q][r] = 1;
    }
  }
  if (use_weight) {
    if((w1_fp = fopen(w1_file, "rt")) == NULL){
      fprintf(stderr, "Can't open %s\n", w1_file);
      exit(-1);
    }
    read_weight_file(w1_fp, weights, nseq);
    fclose(w1_fp);
  }
  if (use_parametric_weight) {
    if ((weights1 = (int **)malloc(nseq*sizeof(int *))) == NULL) {
      fprintf(stderr,"Could not allocate memory for wight matrix\n");
      exit(1);
    }
    if ((weights2 = (int **)malloc(nseq*sizeof(int *))) == NULL) {
      fprintf(stderr,"Could not allocate memory for wight matrix\n");
      exit(1);
    }
    if ((weights3 = (int **)malloc(nseq*sizeof(int *))) == NULL) {
      fprintf(stderr,"Could not allocate memory for wight matrix\n");
      exit(1);
    }
    for (q = 0;q < nseq;q++) {
      if ((weights1[q] = (int *)malloc(nseq*sizeof(int))) == NULL) {
	fprintf(stderr,"Could not allocate memory for wight matrix\n");
	exit(1);
      }
      if ((weights2[q] = (int *)malloc(nseq*sizeof(int))) == NULL) {
	fprintf(stderr,"Could not allocate memory for wight matrix\n");
	exit(1);
      }
      if ((weights3[q] = (int *)malloc(nseq*sizeof(int))) == NULL) {
	fprintf(stderr,"Could not allocate memory for wight matrix\n");
	exit(1);
      }
    }
    if((w1_fp = fopen(w1_file, "rt")) == NULL){
      fprintf(stderr, "Can't open %s\n", w1_file);
      exit(-1);
    }
    if((w2_fp = fopen(w2_file, "rt")) == NULL){
      fprintf(stderr, "Can't open %s\n", w2_file);
      exit(-1);
    }
    read_weight_file(w1_fp, weights1, nseq);
    fclose(w1_fp);/*print_weights(weights1,nseq);*/
    read_weight_file(w2_fp, weights2, nseq);
    fclose(w2_fp);/*print_weights(weights2,nseq);*/
    for (i = 0; i<nseq; i++)
      for (j = 0; j<nseq; j++) weights3[i][j] = weights2[i][j]-weights1[i][j];
  }

  if(use_parametric_score) {
    for (i = 0;i < (('Z'-'A'+1)+2);i++)
      for (j = 0;j < (('Z'-'A'+1)+2);j++)
	score_table3[i][j] = score_table2[i][j] - score_table1[i][j];
  }

  if(use_parametric_weight) {
    printf("Parametric Obsevation between Weight Matrices:\n");
    if(use_group){
      fprintf(stderr,"Not implemented : group Eppstein\n");exit(1);
    }
    if(nseq > MAX_DIM){
      fprintf(stderr, "Using only first %d sequences\n", MAX_DIM);
      dim = MAX_DIM;
    }else{
      dim = nseq;
    }
    print_weights(weights1, dim);
    printf("and\n");
    print_weights(weights2, dim);
    printf("\n");

    for(i = 0; i < dim; i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
    }
    if((dir_mask & FMASK) && (dir_mask & BMASK)){
      fprintf(stderr,"Not implemented : bidirectional for parametric\n");exit(1);
    }
    if(no_ope){
      exit(0);
    }

    for(i=0;i<dim;i++) for(j=0;j<dim;j++)
      weights[i][j]=weights1[i][j];
    boundaries2[0].left = -1;
    boundaries2[0].right = -1;
    boundaries2[0].p = 0;
    boundaries2[0].q = 1;
    pairsubopt(dim, al, use_ub_parametric, -lbound1,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, NULL, (&(boundaries2[0])), NULL);

    for(i=0;i<dim;i++) for(j=0;j<dim;j++)
      weights[i][j]=weights2[i][j];
    boundaries2[1].left = -1;
    boundaries2[1].right = -1;
    boundaries2[1].p = 1;
    boundaries2[1].q = 0;
    pairsubopt(dim, al, use_ub_parametric, -lbound2,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, NULL, (&(boundaries2[1])), NULL);

    iter_parametric_weight(dim, al, k, fs_k, bs_k, dir_mask, line,
             print_subopt, boundaries2, 0, 1, &max_iteration,
             &next_iteration);

    printf("====== RESULTS =====\n");
    print_boundaries2(boundaries2, line, next_iteration, dim);
    exit(0);
  }

  if(use_parametric_score) {
    printf("Parametric Obsevation between Score Matrices:\n");
    if(use_group){
      fprintf(stderr,"Not implemented : group Eppstein\n");exit(1);
    }
    if(nseq > MAX_DIM){
      fprintf(stderr, "Using only first %d sequences\n", MAX_DIM);
      dim = MAX_DIM;
    }else{
      dim = nseq;
    }
    printf(" First Score Table : %s\n",score_file);
    printf(" Second Score Table: %s\n\n",score_file2);

    for(i = 0; i < dim; i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
    }
    if((dir_mask & FMASK) && (dir_mask & BMASK)){
      fprintf(stderr,"Not implemented : bidirectional for parametric\n");exit(1);
    }
    if(no_ope){
      exit(0);
    }

    boundaries3[0].left = -1;
    boundaries3[0].right = -1;
    boundaries3[0].p = 0;
    boundaries3[0].q = 1;
    pairsubopt(dim, al, use_ub_parametric, -lbound1,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, NULL, NULL, (&(boundaries3[0])));

    for(i=0;i<(('Z'-'A'+1)+2);i++) for(j=0;j<(('Z'-'A'+1)+2);j++)
      score_table[i][j]=score_table2[i][j];
    max_score = max_score2;
    internal_gap_penalty = internal_gap_penalty2;
    external_gap_penalty = external_gap_penalty2;
    boundaries3[1].left = -1;
    boundaries3[1].right = -1;
    boundaries3[1].p = 1;
    boundaries3[1].q = 0;
    pairsubopt(dim, al, use_ub_parametric, -lbound2,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, NULL, NULL, (&(boundaries3[1])));

    iter_parametric_score(dim, al, k, fs_k, bs_k, dir_mask, line,
             print_subopt, boundaries3, 0, 1, &max_iteration,
             &next_iteration);

    printf("====== RESULTS =====\n");
    print_boundaries3(boundaries3, line, next_iteration, dim);
    exit(0);
  }

  if(use_parametric_gap) {
    printf("Parametric Obsevation of gap penalty between ");
    printf("%d/%d and %d/%d\n",
	   start_gap_penalty, start_gap_denominator,
	   end_gap_penalty, end_gap_denominator);
    if(use_group){
      fprintf(stderr,"Not implemented : group Eppstein\n");exit(1);
    }
    if(nseq > MAX_DIM){
      fprintf(stderr, "Using only first %d sequences\n", MAX_DIM);
      dim = MAX_DIM;
    }else{
      dim = nseq;
    }
    for(i = 0; i < dim; i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
    }
    if((dir_mask & FMASK) && (dir_mask & BMASK)){
      fprintf(stderr,"Not implemented : bidirectional for parametric\n");exit(1);
    }
    if(no_ope){
      exit(0);
    }

    internal_gap_penalty = start_gap_penalty;
    external_gap_penalty = start_gap_penalty;
    gap_denominator = start_gap_denominator;
    boundaries[0].left = -1;
    boundaries[0].right = -1;
    pairsubopt(dim, al, use_ub_parametric, -lbound1,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, (&(boundaries[0])), NULL, NULL);

    internal_gap_penalty = end_gap_penalty;
    external_gap_penalty = end_gap_penalty;
    gap_denominator = end_gap_denominator;
    boundaries[1].left = -1;
    boundaries[1].right = -1;
    pairsubopt(dim, al, use_ub_parametric, -lbound2,
	       k, fs_k, bs_k, dir_mask, 0, -1,
	       line, 0, print_subopt, (&(boundaries[1])), NULL, NULL);

    if ((start_gap_penalty * end_gap_denominator) < (end_gap_penalty * start_gap_denominator)) {
      iter_parametric_gap(dim, al, k, fs_k, bs_k, dir_mask, line,
           print_subopt, boundaries, 0, 1, &max_iteration,
           &next_iteration);
    } else {
      iter_parametric_gap(dim, al, k, fs_k, bs_k, dir_mask, line,
           print_subopt, boundaries, 1, 0, &max_iteration,
           &next_iteration);
    }

    printf("====== RESULTS =====\n");
    print_boundaries(boundaries, line, next_iteration);
    exit(0);
  }


  if(use_k|use_delta) {
    if(!min_output){
      printf("Compute suboptimal alignments with Eppstein and A* (");
      if((dir_mask & FMASK) && (dir_mask & BMASK)){
	printf("bidirectional ");
      }else if(dir_mask & FMASK){
	printf("forward ");
      }else if(dir_mask & BMASK){
	printf("backward ");
      }
      printf("search):\n");
      if(use_ub) printf("  Upper bound: %d\n", -lbound);
      printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
      printf("  Delta: %d\n", delta);
      if(use_k) {
	printf("  #solutions : %d\n", kbs);
	fprintf(stderr," k best solutions are not implemented\n");
      }
    }
    if(use_group){
      fprintf(stderr,"Not implemented : group Eppstein\n");exit(1);
    }
    if(nseq > MAX_DIM){
      fprintf(stderr, "Using only first %d sequences\n", MAX_DIM);
      dim = MAX_DIM;
    }else{
      dim = nseq;
    }
    for(i = 0; i < dim; i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
    }
    if((dir_mask & FMASK) && (dir_mask & BMASK)){
      fprintf(stderr,"Not implemented : bidirectional Eppstein\n");exit(1);
    }
    if(no_ope){
      exit(0);
    }
    pairsubopt(dim, al, use_ub, -lbound, k, fs_k, bs_k, dir_mask, delta , kbs, line, use_detour, print_subopt, NULL, NULL, NULL);
    exit(0);
  }

  if(use_group){
    if(!min_output){
      printf("Align incremantally (");
      if((dir_mask & FMASK) && (dir_mask & BMASK)){
	printf("bidirectional ");
      }else if(dir_mask & FMASK){
	printf("forward ");
      }else if(dir_mask & BMASK){
	printf("backward ");
      }
      printf("search):\n");
      printf("  # of groups: %d\n", ngroup);
      printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
    }
    if(no_ope){
      exit(0);
    }
    if(ngroup < 2){
      fprintf(stderr, "# of groups is too small: %d\n", ngroup);
      exit(-1);
    }
    dim = 0;
    for(i = 0; (i < nseq) && (i < ngroup); i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
      dim++;
    }
    if(dim == 2){
      result_al_p = dp(2, al);
    }else if(dim > 2){
      result_al_p = astar_pairopt(dim, al, k, fs_k, bs_k, dir_mask);
    }
    for(i = dim; i < nseq; i += ngroup - 1){
      al[0] = *result_al_p;
      dim = 1;
      for(j = 0; (i + j < nseq) && (j < ngroup - 1); j++){
	al[j + 1].dim = 1;
	al[j + 1].seq = &(seq[i + j]);
	dim++;
      }
      if(dim == 2){
	result_al_p = dp(2, al);
      }else if(dim > 2){
	result_al_p = astar_pairopt(dim, al, k, fs_k, bs_k, dir_mask);
      }
      for(j = 0; j < al[0].dim; j++){
	free(al[0].seq[j].str);
      }
      free(al[0].seq);
    }
  }else{
    if(nseq > MAX_DIM){
      fprintf(stderr, "Using only first %d sequences\n", MAX_DIM);
      dim = MAX_DIM;
    }else{
      dim = nseq;
    }
    for(i = 0; i < dim; i++){
      al[i].dim = 1;
      al[i].seq = &(seq[i]);
    }
    if(use_astar){
      if(use_pairopt){
	if(use_ub){
	  if(!min_output){
	    printf("Align with A* utilizing upper bound (");
	    if((dir_mask & FMASK) && (dir_mask & BMASK)){
	      printf("bidirectional ");
	    }else if(dir_mask & FMASK){
	      printf("forward ");
	    }else if(dir_mask & BMASK){
	      printf("backward ");
	    }
	    printf("search):\n");
	    printf("  Upper bound: %d\n", -lbound);
	    printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	  }
	  if(no_ope){
	    exit(0);
	  }
	  result_al_p
	    = astar_pairopt_ub(dim, al, -lbound, k, fs_k, bs_k, dir_mask);
	}else if(use_step){
	  if(!min_output){
	    printf("Align with A* expanding one sequence at a time (");
	    if((dir_mask & FMASK) && (dir_mask & BMASK)){
	      printf("bidirectional ");
	    }else if(dir_mask & FMASK){
	      printf("forward ");
	    }else if(dir_mask & BMASK){
	      printf("backward ");
	    }
	    printf("search):\n");
	    printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	  }
	  if(no_ope){
	    exit(0);
	  }
	  result_al_p
	    = astar_pairopt_step(dim, al, k, fs_k, bs_k, dir_mask);
/*
        }else if(use_inc){
	  if(!min_output){
	    printf("Align with A* inrementing upper bound (");
	    if((dir_mask & FMASK) && (dir_mask & BMASK)){
	      printf("bidirectional ");
	    }else if(dir_mask & FMASK){
	      printf("forward ");
	    }else if(dir_mask & BMASK){
	      printf("backward ");
	    }
	    printf("search):\n");
	    printf("  Initial delta: %d\n", dlt);
	    printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	  }
	  if(no_ope){
	    exit(0);
	  }
          result_al_p
	    = astar_pairopt_inc(dim, al, dlt, k, fs_k, bs_k, dir_mask);
	}else if(use_accel){
	  if(!min_output){
	    printf("Align with A using the number of edges in a path (");
	    if((dir_mask & FMASK) && (dir_mask & BMASK)){
	      printf("bidirectional ");
	    }else if(dir_mask & FMASK){
	      printf("forward ");
	    }else if(dir_mask & BMASK){
	      printf("backward ");
	    }
	    printf("search):\n");
	    printf("  Coefficient: %d\n", accel);
	    printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	  }
	  if(no_ope){
	    exit(0);
	  }
	  result_al_p
	    = astar_pairopt_accel(dim, al, accel, k, fs_k, bs_k, dir_mask);
*/
	}else{
	  if(!min_output){
	    printf("Align with A* (");
	    if((dir_mask & FMASK) && (dir_mask & BMASK)){
	      printf("bidirectional ");
	    }else if(dir_mask & FMASK){
	      printf("forward ");
	    }else if(dir_mask & BMASK){
	      printf("backward ");
	    }
	    printf("search):\n");
	    printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	  }
	  if(no_ope){
	    exit(0);
	  }
	  result_al_p = astar_pairopt(dim, al, k, fs_k, bs_k, dir_mask);
	}
      }else{
	if(!min_output){
	  printf("Align with A* using estimators based on the score table(");
	  if((dir_mask & FMASK) && (dir_mask & BMASK)){
	    printf("bidirectional ");
	  }else if(dir_mask & FMASK){
	    printf("forward ");
	  }else if(dir_mask & BMASK){
	    printf("backward ");
	  }
	  printf("search):\n");
	  printf("  k: %d/%d, %d/%d\n", fs_k, k, bs_k, k);
	}
	if(no_ope){
	  exit(0);
	}
	result_al_p = astar(dim, al, k, fs_k, bs_k, dir_mask);
      }
    }else{
      if(!min_output){
	printf("Align with DP:\n");
      }
      if(no_ope){
	exit(0);
      }
      result_al_p = dp(dim, al);
    }
  }
#ifdef VERBOSE
  t_end = time(NULL);
#endif
  if(min_output){
    for(i = 0; i < result_al_p->dim; i++){
      for(j = 0; j < result_al_p->seq[0].len; j++){
	if(result_al_p->seq[i].str[j] == INTERNAL_GAP_CHAR){
	  putchar('-');
	}else if(result_al_p->seq[i].str[j] == EXTERNAL_GAP_CHAR){
	  putchar('=');
	}else{
	  putchar(result_al_p->seq[i].str[j]);
	}
      }
      printf("\n");
    }
  }else{
#ifdef VERBOSE
    fprintf(stderr, "Actually used time: %d (sec.)\n",
	    (long)t_end - (long)t_start);
#endif
    printf("Alignment:");
    for(i = 0; i < result_al_p->seq[0].len; i += line){
      printf("\n");
      printf("  %-24d  %24d\n", i + 1, i + line);
      for(j = 0; j < result_al_p->dim; j++){
	printf("  ");
	for(n = 0; n < line; n++){
	  if(i + n >= result_al_p->seq[0].len){
	    break;
	  }
	  if(result_al_p->seq[j].str[i + n] == INTERNAL_GAP_CHAR){
	    putchar('-');
	  }else if(result_al_p->seq[j].str[i + n] == EXTERNAL_GAP_CHAR){
	    putchar('=');
	  }else{
	    putchar(result_al_p->seq[j].str[i + n]);
	  }
	}
	printf("\n");
      }
    }
    printf("Score: %d\n", result_al_p->score);
  }
}
