#ifndef _ALIGN_H

#define _ALIGN_H

#if defined(linux)
#define SAVE_MEMORY
#endif

#define MAX_DIM 32
#define MAX_NSEQ 100
#define MAXSOLUTIONS 64
#define BUFSIZE 128

enum {FORWARD, BACKWARD};

#define FMASK (1 << FORWARD)
#define BMASK (1 << BACKWARD)

enum{UNVISITED, VISITED, FINISHED, CONSTRUCTED, BESTPATH};

typedef struct node *node_t;
typedef struct epp_node *epp_node_t;
typedef struct edge_heap *edge_heap_t;
typedef struct subopt_answer *subopt_answer_t;

typedef struct node_info{
  int status;
  node_t prev_node;
  int cost;
  int pqueue_index;
  int edge_heap_size;          /* size of H_edge */
  edge_heap_t edge_heap;       /* H_edge */
  epp_node_t tree_heap;        /* H_t */
} *node_info_t;

node_info_t node_info_init(void);
void node_info_finish(node_info_t);
epp_node_t init_epp_node(void);

struct node{
  int *pos;
  node_info_t info[2];
};

struct epp_node{
  node_t node;
  epp_node_t left;
  epp_node_t right;
};

struct edge_heap{
  int delta;
  node_t node;
};

struct subopt_answer{
  node_t from;
  node_t to;
};

node_t node_init(int, int *);
void node_finish(node_t);

#define fs_info_init(node) ((node)->info[FORWARD] = node_info_init())
#define bs_info_init(node) ((node)->info[BACKWARD] = node_info_init())

#define fs_info_reset(node) ((node)->info[FORWARD]->status = UNVISITED)
#define bs_info_reset(node) ((node)->info[BACKWARD]->status = UNVISITED)

#define fs_info(node) ((node)->info[FORWARD])
#define bs_info(node) ((node)->info[BACKWARD])

#define is_fs_visited(node) \
  ((fs_info(node) != NULL) && (fs_info(node)->status == VISITED))
#define is_fs_finished(node) \
  ((fs_info(node) != NULL) && (fs_info(node)->status == FINISHED))
#define is_bs_visited(node) \
  ((bs_info(node) != NULL) && (bs_info(node)->status == VISITED))
#define is_bs_finished(node) \
  ((bs_info(node) != NULL) && (bs_info(node)->status == FINISHED))
#define is_not_finished(node, dir) \
  ((((node)->info[dir]) != NULL) && (((node)->info[dir])->status < FINISHED))
#define is_finished(node, dir) \
  ((((node)->info[dir]) != NULL) && (((node)->info[dir])->status >= FINISHED))
#define is_constructed(node, dir) \
  ((((node)->info[dir]) != NULL) && (((node)->info[dir])->status >= CONSTRUCTED))
#define is_on_the_best_path(node, dir) \
  ((((node)->info[(dir)]) != NULL) && (((node)->info[(dir)])->status == BESTPATH))

#define set_status(node, d, s) ((node)->info[d]->status = (s))
#define set_prev_node(node, d, p) ((node)->info[d]->prev_node = (p))
#define set_cost(node, d, c) ((node)->info[d]->cost = (c))
#define set_pqueue_index(node, d, i) ((node)->info[d]->pqueue_index = (i)) 
#define lookup_edge_heap_size(node, d) ((node)->info[d]->edge_heap_size)
#define set_edge_heap_size(node, d, sz) ((node)->info[d]->edge_heap_size = (sz))
#define set_edge_heap(node, d, he) ((node)->info[d]->edge_heap = (he))
#define lookup_edge_heap(node, d) ((node)->info[d]->edge_heap)
#define malloc_edge_heap(node, d, sz) set_edge_heap(node, d,\
                    (edge_heap_t)malloc((sz)*sizeof(struct edge_heap)))
#define realloc_edge_heap(node, d, sz) set_edge_heap(node, d,\
                       (edge_heap_t)realloc(lookup_edge_heap(node, d),\
                                  (sz)*sizeof(struct edge_heap)))

#define set_delta_of_edge_heap(node,d,i,dlt) (lookup_edge_heap(node, d)[(i)].delta = (dlt))
#define set_node_of_edge_heap(node,d,i,pn)  (lookup_edge_heap(node, d)[(i)].node = (pn))
#define lookup_tree_heap(node, d) (((node)->info[d])->tree_heap)
#define set_tree_heap(node, d, ht) ((node)->info[d]->tree_heap = (ht))

#define set_fs_status(node, s) set_status(node, FORWARD, s)
#define set_fs_prev_node(node, p) set_prev_node(node, FORWARD, p)
#define set_fs_cost(node, c) set_cost(node, FORWARD, c)
#define set_fs_pqueue_index(node, i) set_pqueue_index(node, FORWARD, i)

#define lookup_ht_delta(ht, d) ((lookup_edge_heap((ht)->node, (d)))[0].delta)
#define lookup_ht_node(ht, d)  ((lookup_edge_heap((ht)->node, (d)))[0].node)

#define set_bs_status(node, s) set_status(node, BACKWARD, s)
#define set_bs_prev_node(node, p) set_prev_node(node, BACKWARD, p)
#define set_bs_cost(node, c) set_cost(node, BACKWARD, c)
#define set_bs_pqueue_index(node, i) set_pqueue_index(node, BACKWARD, i)

#define init_subopt_answer(sa, d) ((sa) = (subopt_answer_t)malloc((d)*sizeof(struct subopt_answer)))
#define realloc_subopt_answer(sa, d) ((sa) = (subopt_answer_t)realloc((sa), (d)*sizeof(struct subopt_answer)))
#define set_f_of_subopt_answer(sa, d, node) (((sa)[(d)]).from = (node))
#define set_t_of_subopt_answer(sa, d, node) (((sa)[(d)]).to = (node))
#define lookup_f_of_subopt_answer(sa, d) (((sa)[(d)]).from)
#define lookup_t_of_subopt_answer(sa, d) (((sa)[(d)]).to)

#define lookup_status(node, d) ((node)->info[d]->status)
#define lookup_prev_node(node, d) ((node)->info[d]->prev_node)
#define lookup_cost(node, d) ((node)->info[d]->cost)
#define lookup_pqueue_index(node, d) ((node)->info[d]->pqueue_index)

#define lookup_fs_status(node) lookup_status(node, FORWARD)
#define lookup_fs_prev_node(node) lookup_prev_node(node, FORWARD)
#define lookup_fs_cost(node) lookup_cost(node, FORWARD)
#define lookup_fs_pqueue_index(node) lookup_pqueue_index(node, FORWARD)

#define lookup_bs_status(node) lookup_status(node, BACKWARD)
#define lookup_bs_prev_node(node) lookup_prev_node(node, BACKWARD)
#define lookup_bs_cost(node) lookup_cost(node, BACKWARD)
#define lookup_bs_pqueue_index(node) lookup_pqueue_index(node, BACKWARD)

typedef struct node_list *node_list_t;

struct node_list{
  node_t node;
  node_list_t next;
};

void mul_score();
void div_score();
void init_scores();
int normalize_score();

#define node_list_init() (NULL)
void node_list_finish(node_list_t);
node_list_t node_list_cons(node_t, node_list_t);
node_list_t node_list_decons(node_list_t);

typedef struct node_set{
  int dim;
  int count;
  int table_size;
#ifdef DEBUG
  int long_list_len;
#endif
  node_list_t *node_list_table;
} *node_set_t;

node_set_t node_set_init(int);
void node_set_finish(node_set_t);
void node_set_reset(node_set_t);
node_t get_node(node_set_t, int *);
node_t create_node(node_set_t, int *);

typedef void *pqueue_t;

pqueue_t pqueue_init(int);
void pqueue_finish(pqueue_t);
void put_node_to_pqueue(pqueue_t, node_t);
node_t get_node_from_pqueue(pqueue_t);
void update_pqueue(pqueue_t, node_t);
void print_pqueue(pqueue_t);
void extend_node(pqueue_t, node_t, node_t, int);

typedef struct sequence{
  int len;
  char *str;
} sequence_t;

typedef struct alignment{
  int score;
  int dim;
  sequence_t *seq;
} alignment_t;

typedef struct solutions{
  int left;
  int right;
  int gap_penalty;
  int gap_denominator;
  int score;
  alignment_t *max_alignment;
  int max_gap;
  alignment_t *min_alignment;
  int min_gap;
  int number_of_sols;  /* number of optimal solutions */
  int number_of_max_sols;
  int number_of_min_sols;
} solutions_t, *solutions_p;

typedef struct solutions2{
  int left;
  int right;
  int p;
  int q;   /*  w1 -p- w -q- w2 */
  int score;
  alignment_t *max_alignment;
  int max_slope;
  alignment_t *min_alignment;
  int min_slope;
  int number_of_sols;  /* number of optimal solutions */
  int number_of_max_sols;
  int number_of_min_sols;
} solutions2_t, *solutions2_p;

void print_alignment(alignment_t *, int);
void print_number_of_gaps(alignment_t *, int *, int *, int *, int *, int);

void print_subopt_answer(subopt_answer_t, int, node_t, node_t, node_t,
			 int, int, alignment_t *, int, int,
			 solutions_p, solutions2_p, solutions2_p);
void print_boundaries(solutions_p, int, int);
void print_boundaries2(solutions2_p, int, int, int);
void print_boundaries3(solutions2_p, int, int, int);
int takeword(FILE *, char *);
void read_weight_file(FILE *, int **, int);
void print_weights(int **, int);


node_t lookup_sub_fs_prev_node(node_t, subopt_answer_t, int *, int);
node_t lookup_sub_bs_prev_node(node_t, subopt_answer_t, int *, int);
node_t lookup_sub_fs_prev_node_s(node_t, subopt_answer_t, int *, int);
node_t lookup_sub_bs_prev_node_s(node_t, subopt_answer_t, int *, int);

void alloc_alignment(alignment_t *, int, int);
void free_alignment(alignment_t *, int);


int score_table[('Z' + 1 - 'A') + 2][('Z' +  1 - 'A') + 2], max_score;
int score_table1[('Z' + 1 - 'A') + 2][('Z' +  1 - 'A') + 2], max_score1;
int score_table2[('Z' + 1 - 'A') + 2][('Z' +  1 - 'A') + 2], max_score2;
int score_table3[('Z' + 1 - 'A') + 2][('Z' +  1 - 'A') + 2];
int internal_gap_penalty, external_gap_penalty, gap_denominator;
int internal_gap_penalty1, external_gap_penalty1, gap_denominator1;
int internal_gap_penalty2, external_gap_penalty2, gap_denominator2;
int **weights, **weights1, **weights2, **weights3;

#define INTERNAL_GAP_CHAR ((char)('Z' + 1))
#define EXTERNAL_GAP_CHAR ((char)('Z' + 2))

#define is_normal_char(c) ((c) <= 'Z')
#define is_gap_char(c) ((c) > 'Z')
#define matching_cost(x, y) (-score_table[(x) - 'A'][(y) - 'A'])

alignment_t *dp(int, alignment_t *);
alignment_t *astar(int, alignment_t *, int, int, int, int);

#ifdef SIMPLE_EST_TABLE

typedef struct estimator{
  int dir;
  alignment_t *col_al_p;
  alignment_t *row_al_p;
  int *table;
} *estimator_t;

estimator_t estimator_init(alignment_t *, alignment_t *, int);
void estimator_finish(estimator_t);

#define set_estimate(estimator, i, j, estimate) \
  ((estimator)->table[(i) * ((estimator)->row_al_p->seq[0].len  + 1) + (j)] \
   = (estimate))
#define lookup_estimate(estimator, i, j) \
  ((estimator)->table[(i) * ((estimator)->row_al_p->seq[0].len  + 1) + (j)])

#else

typedef struct estimator{
  int dir;
  alignment_t *col_al_p;
  alignment_t *row_al_p;
  int start;
  int end;
  int **table;
} *estimator_t;

estimator_t estimator_init(alignment_t *, alignment_t *, int);
void estimator_finish(estimator_t);
void set_estimate(estimator_t, int, int, int);
int lookup_estimate(estimator_t, int, int);

#endif

alignment_t *astar_pairopt(int, alignment_t *, int, int, int, int);
alignment_t *astar_pairopt_accel(int, alignment_t *, int, int, int, int, int);
alignment_t *astar_pairopt_ub(int, alignment_t *, int, int, int, int, int);
alignment_t *astar_pairopt_inc(int, alignment_t *, int, int, int, int, int);
alignment_t *astar_pairopt_step(int, alignment_t *, int, int, int, int);

void pairsubopt(int, alignment_t *, int, int, int, int, int, int, int, int, int, int, int, solutions_p, solutions2_p, solutions2_p);
void construct_heap(node_t, node_set_t, int, alignment_t *, int, int, int, int, int,
#ifdef VERBOSE
		    int *, int *, int *,
#endif
		    int, estimator_t *, int);

void tune_h_edge_down(edge_heap_t, int, int);
void swap_h_edge(edge_heap_t, int, int);
void construct_tree_heap(
#ifdef VERBOSE
			 int *,
#endif
			 node_t, node_t, int);
void swap_epp_node(epp_node_t, epp_node_t);
void make_histogram_for_h_t(epp_node_t, int, int *,
			    node_t, node_t, node_t,
			    node_set_t, int, alignment_t *,
			    int, int, int, int, int,
#ifdef VERBOSE
			    int *, int *, int *,
#endif
			    int, estimator_t *,
			    int, subopt_answer_t, int, int *, int,
			    solutions_p, solutions2_p, solutions2_p);
void make_histogram_for_h_t_d1(epp_node_t, int, int *,
			       node_t, node_t, node_t,
			       node_set_t, int, alignment_t *,
			       int, int, int, int, int,
#ifdef VERBOSE
			       int *, int *, int *,
#endif
			       int, estimator_t *,
			       int, subopt_answer_t, int, int *, int);
void make_histogram_for_h_t_d2(epp_node_t, int, int *,
			       node_t,  node_t, node_t,
			       node_set_t, int, alignment_t *,
			       int, int, int, int, int,
#ifdef VERBOSE
			       int *, int *, int *,
#endif
			       int, estimator_t *,
			       int, subopt_answer_t, int, int *, int);

void make_histogram_for_h_edge(edge_heap_t, node_t, int, int, int, int *,
			       node_t, node_t, node_t,
			       node_set_t, int, alignment_t *,
			       int, int, int, int, int,
#ifdef VERBOSE
			       int *, int *, int *,
#endif
			       int, estimator_t *,
			       int, subopt_answer_t, int, int *, int,
			       solutions_p, solutions2_p, solutions2_p);
void make_histogram_for_h_edge_d(edge_heap_t, node_t, int, int, int, int *,
				 node_t, node_t, node_t,
				 node_set_t, int, alignment_t *,
				 int, int, int, int, int,
#ifdef VERBOSE
				 int *, int *, int *,
#endif
				 int, estimator_t *,
				 int, subopt_answer_t, int, int *, int);

void iter_parametric_gap(int, alignment_t *, int, int, int, int, int,
   int, solutions_p, int, int, int *, int *);
void iter_parametric_weight(int, alignment_t *, int, int, int, int, int,
   int, solutions2_p, int, int, int *, int *);
void iter_parametric_score(int, alignment_t *, int, int, int, int, int,
   int, solutions2_p, int, int, int *, int *);

alignment_t *iter_random(alignment_t *, int, int, int, int, int, int);
int gcd_primitive(int, int);
int gcd(int, int);
int normalize_weights(int **, int);
int weighted_score(alignment_t *, int **);
int weighted_score_gd(alignment_t *, int **);
int score_slope(alignment_t *);
int another_score1(alignment_t *);
int another_score2(alignment_t *);

#endif
