#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

alignment_t *astar(int dim, alignment_t *alignment,
		   int wk, int fs_k, int bs_k, int dir_mask)
{
  pqueue_t fs_pqueue, bs_pqueue;
  node_set_t node_set;
  node_t node, next_node, mid_node, orig_node, dest_node;
  node_list_t bridge_node_list, node_list;
  int *fmin_cost['Z' + 1 - 'A' + 2][MAX_DIM];
  int *bmin_cost['Z' + 1 - 'A' + 2][MAX_DIM];
  int *tmp_table, edge_cost, add_cost, est_cost, est_diff;
  int i, j, k, l, m, n, p, q, is_bridge;
  int al_len, al_score;
  int next_pos[MAX_DIM], pos[MAX_DIM];
  int diag_pos, diag_next_pos;
  alignment_t *al_p;
#ifdef VERBOSE
  double wnodes;
  int fs_visited, bs_visited, fs_expanded, bs_expanded;
  struct tms t_start, t_search_start, t_search_end, t_end;
#endif

#ifdef VERBOSE
  times(&t_start);
#endif
  if(dim <= 0){
    return NULL;
  }else if(dim == 1){
    return alignment;
  }
  n = 0;
  for(i = 0; i < dim; i++){
/*    if(alignment[i].seq[0].len > n){*/
    if(alignment[i].dim > n){
      n = alignment[i].dim;
    }
  }
  add_cost = n * n * max_score;
#ifdef VERBOSE
  fprintf(stderr, "Calculating estimates .... ");
#endif
  for(i = 'A'; i <= 'Z' + 2; i++){
    for(j = 0; j < dim; j++){
      if((tmp_table
	  = (int *)calloc(alignment[j].seq[0].len, sizeof(int))) == NULL){
	fprintf(stderr, "Can't allocate memmory for estimates\n");
	exit(-1);
      }
      for(k = 0; k < alignment[j].dim; k++){
	for(l = 0; l < alignment[j].seq[0].len; l++){
	  tmp_table[l] += matching_cost(i, alignment[j].seq[k].str[l]);
	}
      }
      if(dir_mask & FMASK){
	if(dir_mask & BMASK){
	  if((fmin_cost[i - 'A'][j]
	      = (int *)malloc(sizeof(int) * alignment[j].seq[0].len)) == NULL){
	    fprintf(stderr, "Can't allocate memmory for estimates\n");
	    exit(-1);
	  }
	  fmin_cost[i - 'A'][j][0] = tmp_table[0];
	  for(k = 1; k < alignment[j].seq[0].len; k++){
	    if(fmin_cost[i - 'A'][j][k - 1] < tmp_table[k]){
	      fmin_cost[i - 'A'][j][k] = fmin_cost[i - 'A'][j][k - 1];
	    }else{
	      fmin_cost[i - 'A'][j][k] = tmp_table[k];
	    }
	  }
	}else{
	  for(k = 1; k < alignment[j].seq[0].len; k++){
	    if(tmp_table[k - 1] < tmp_table[k]){
	      tmp_table[k] = tmp_table[k - 1];
	    }
	  }
	  fmin_cost[i - 'A'][j] = tmp_table;
	}
      }
      if(dir_mask & BMASK){
	for(k = alignment[j].seq[0].len - 2; k >= 0; k--){
	  if(tmp_table[k + 1] < tmp_table[k]){
	    tmp_table[k] = tmp_table[k + 1];
	  }
	}
	bmin_cost[i - 'A'][j] = tmp_table;
      }
    }
  }
#ifdef VERBOSE
  fprintf(stderr, "done\n");
  times(&t_search_start);
  fs_visited = 0;
  bs_visited = 0;
  fs_expanded = 0;
  bs_expanded = 0;
#endif
  node_set = node_set_init(dim);
  for(i = 0; i < dim; i++){
    pos[i] = 0;
  }
  orig_node = create_node(node_set, pos);
  fs_info_init(orig_node);
  for(i = 0; i < dim; i++){
    pos[i] = alignment[i].seq[0].len;
  }
  dest_node = create_node(node_set, pos);
  bs_info_init(dest_node);
  bridge_node_list = node_list_init();
  if(dir_mask & FMASK){
    fs_pqueue = pqueue_init(FORWARD);
    est_cost = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	if(external_gap_penalty > internal_gap_penalty){
	  if(alignment[i].seq[0].len < alignment[j].seq[0].len){
	    est_cost
	      += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		[j][alignment[j].seq[0].len - 1]
		  * (alignment[j].seq[0].len - alignment[i].seq[0].len);
	  }else{
	    est_cost
	      += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		[i][alignment[i].seq[0].len - 1]
		  * (alignment[i].seq[0].len - alignment[j].seq[0].len);
	  }
	}else{
	  if(alignment[i].seq[0].len < alignment[j].seq[0].len){
	    est_cost
	      += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		[j][alignment[j].seq[0].len - 1]
		  * (alignment[j].seq[0].len - alignment[i].seq[0].len);
	  }else{
	    est_cost
	      += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		[i][alignment[i].seq[0].len - 1]
		  * (alignment[i].seq[0].len - alignment[j].seq[0].len);
	  }
	}
	for(k = 0; (k < alignment[i].seq[0].len)
	    && (k < alignment[j].seq[0].len); k++){
	  if(external_gap_penalty > internal_gap_penalty){
	    est_diff = fmin_cost[EXTERNAL_GAP_CHAR - 'A']
	      [j][alignment[j].seq[0].len - 1 - k]
		* alignment[i].dim
		  + fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		    [i][alignment[i].seq[0].len - 1 - k]
		      * alignment[j].dim;
	  }else{
	    est_diff = fmin_cost[INTERNAL_GAP_CHAR - 'A']
	      [j][alignment[j].seq[0].len - 1 - k]
		* alignment[i].dim
		  + fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [i][alignment[i].seq[0].len - 1 - k]
		      * alignment[j].dim;
	  }
	  n = 0;
	  for(l = 0; l < alignment[i].dim; l++){
	    n += fmin_cost[alignment[i].seq[l]
			   .str[alignment[i].seq[0].len - 1 - k] - 'A']
			     [j][alignment[j].seq[0].len - 1 - k];
	  }
	  if(n < est_diff){
	    est_diff = n;
	  }
	  n = 0;
	  for(m = 0; m < alignment[j].dim; m++){
	    n += fmin_cost[alignment[j].seq[m]
			   .str[alignment[j].seq[0].len - 1 - k] - 'A']
			     [i][alignment[i].seq[0].len - 1 - k];
	  }
	  if(n < est_diff){
	    est_diff = n;
	  }
	  est_cost += fs_k * est_diff;
	}
	est_cost += fs_k
	  * add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len);
      }
    }
    set_fs_cost(orig_node, est_cost);
    set_fs_status(orig_node, VISITED);
    put_node_to_pqueue(fs_pqueue, orig_node);
#ifdef VERBOSE
    fs_visited++;
#endif
  }else{
    set_fs_cost(orig_node, 0);
  }
  if(dir_mask & BMASK){
    bs_pqueue = pqueue_init(BACKWARD);
    est_cost = 0;
    for(i = 0; i < dim; i++){
      for(j = i + 1; j < dim; j ++){
	if(external_gap_penalty > internal_gap_penalty){
	  if(alignment[i].seq[0].len < alignment[j].seq[0].len){
	    est_cost
	      += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		* (alignment[j].seq[0].len - alignment[i].seq[0].len);
	  }else{
	    est_cost
	      += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][i][0]
		* (alignment[i].seq[0].len - alignment[j].seq[0].len);
	  }
	}else{
	  if(alignment[i].seq[0].len < alignment[j].seq[0].len){
	    est_cost
	      += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		* (alignment[j].seq[0].len - alignment[i].seq[0].len);
	  }else{
	    est_cost
	      += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][i][0]
		* (alignment[i].seq[0].len - alignment[j].seq[0].len);
	  }
	}
	for(k = 0; (k < alignment[i].seq[0].len)
	    && (k < alignment[j].seq[0].len); k++){
	  if(external_gap_penalty > internal_gap_penalty){
	    est_diff = bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][k]
	      * alignment[i].dim
		+ bmin_cost[EXTERNAL_GAP_CHAR - 'A'][i][k] * alignment[j].dim;
	  }else{
	    est_diff = bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][k]
	      * alignment[i].dim
		+ bmin_cost[INTERNAL_GAP_CHAR - 'A'][i][k] * alignment[j].dim;
	  }
	  n = 0;
	  for(l = 0; l < alignment[i].dim; l++){
	    n += bmin_cost[alignment[i].seq[l].str[k] - 'A'][j][k];
	  }
	  if(n < est_diff){
	    est_diff = n;
	  }
	  n = 0;
	  for(m = 0; m < alignment[j].dim; m++){
	    n += bmin_cost[alignment[j].seq[m].str[k] - 'A'][i][k];
	  }
	  if(n < est_diff){
	    est_diff = n;
	  }
	  est_cost += bs_k * est_diff;
	}
	est_cost += bs_k
	  * add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len);
      }
    }
    set_bs_cost(dest_node, est_cost);
    set_bs_status(dest_node, VISITED);
    put_node_to_pqueue(bs_pqueue, dest_node);
#ifdef VERBOSE
    bs_visited++;
#endif
  }else{
    set_bs_cost(dest_node, 0);
  }
  while(1){
    if(dir_mask & FMASK){
      node = get_node_from_pqueue(fs_pqueue);
      set_fs_status(node, FINISHED);
#ifdef VERBOSE
      fs_expanded++;
#endif
      if((node == dest_node) || is_bs_finished(node)){
	mid_node = node;
	break;
      }
      is_bridge = 0;
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] + 1) > alignment[j].seq[0].len){
	      goto fs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  edge_cost += wk
		    * matching_cost(alignment[j].seq[l].str[node->pos[j]],
				    alignment[k].seq[m].str[node->pos[k]]);
		}
	      }
	      edge_cost += wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[node->pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((node->pos[k] == 0)
		    || (alignment[k].seq[m].str[node->pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[node->pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[node->pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((node->pos[j] == 0)
		    || (alignment[j].seq[l].str[node->pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((node->pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[node->pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }
	    if(alignment[j].seq[0].len - node->pos[j]
	       < alignment[k].seq[0].len - node->pos[k]){
	      diag_pos = alignment[j].seq[0].len - node->pos[j];
	    }else{
	      diag_pos = alignment[k].seq[0].len - node->pos[k];
	    }
	    if(alignment[j].seq[0].len - next_pos[j]
	       < alignment[k].seq[0].len - next_pos[k]){
	      diag_next_pos = alignment[j].seq[0].len - next_pos[j];
	    }else{
	      diag_next_pos = alignment[k].seq[0].len - next_pos[k];
	    }
	    if(diag_pos != diag_next_pos){
	      if(external_gap_penalty > internal_gap_penalty){
		est_diff = fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		  [k][alignment[k].seq[0].len - diag_pos] * alignment[j].dim
		    + fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - diag_pos]
			* alignment[k].dim;
	      }else{
		est_diff = fmin_cost[INTERNAL_GAP_CHAR - 'A']
		  [k][alignment[k].seq[0].len - diag_pos] * alignment[j].dim
		    + fmin_cost[INTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - diag_pos]
			* alignment[k].dim;
	      }
	      n = 0;
	      for(l = 0; l < alignment[j].dim; l++){
		n += fmin_cost[alignment[j].seq[l]
			       .str[alignment[j].seq[0].len - diag_pos] - 'A']
				 [k][alignment[k].seq[0].len - diag_pos];
	      }
	      if(n < est_diff){
		est_diff = n;
	      }
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		n += fmin_cost[alignment[k].seq[m]
			       .str[alignment[k].seq[0].len - diag_pos] - 'A']
				 [j][alignment[j].seq[0].len - diag_pos];
	      }
	      if(n < est_diff){
		est_diff = n;
	      }
	      edge_cost -= fs_k * est_diff;
	    }
	    if(external_gap_penalty > internal_gap_penalty){
	      if(alignment[j].seq[0].len - node->pos[j]
		 < alignment[k].seq[0].len - node->pos[k]){
		edge_cost
		  -= fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - 1]
		      * ((alignment[k].seq[0].len - node->pos[k])
			 - (alignment[j].seq[0].len - node->pos[j]));
	      }else{
		  edge_cost
		    -= fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - 1]
			* ((alignment[j].seq[0].len - node->pos[j])
			   - (alignment[k].seq[0].len - node->pos[k]));
		}
	      if(alignment[j].seq[0].len - next_pos[j]
		 < alignment[k].seq[0].len - next_pos[k]){
		edge_cost
		  += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - 1]
		      * ((alignment[k].seq[0].len - next_pos[k])
			 - (alignment[j].seq[0].len - next_pos[j]));
	      }else{
		edge_cost
		  += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		    [j][alignment[j].seq[0].len - 1]
		      * ((alignment[j].seq[0].len - next_pos[j])
			 - (alignment[k].seq[0].len - next_pos[k]));
	      }
	    }else{
	      if(alignment[j].seq[0].len - node->pos[j]
		 < alignment[k].seq[0].len - node->pos[k]){
		edge_cost
		  -= fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - 1]
		      * ((alignment[k].seq[0].len - node->pos[k])
			 - (alignment[j].seq[0].len - node->pos[j]));
	      }else{
		edge_cost
		  -= fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [j][alignment[j].seq[0].len - 1]
		      * ((alignment[j].seq[0].len - node->pos[j])
			 - (alignment[k].seq[0].len - node->pos[k]));
	      }
	      if(alignment[j].seq[0].len - next_pos[j]
		 < alignment[k].seq[0].len - next_pos[k]){
		edge_cost
		  += fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - 1]
		      * ((alignment[k].seq[0].len - next_pos[k])
			 - (alignment[j].seq[0].len - next_pos[j]));
	      }else{
		edge_cost
		  += fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [j][alignment[j].seq[0].len - 1]
		      * ((alignment[j].seq[0].len - next_pos[j])
			 - (alignment[k].seq[0].len - next_pos[k]));
	      }
	    }
	    edge_cost -= fs_k * add_cost * ((next_pos[j] - node->pos[j])
					    + (next_pos[k] - node->pos[k]));
	    if(dir_mask & BMASK){
	      if(node->pos[j] < node->pos[k]){
		diag_pos = node->pos[j];
	      }else{
		diag_pos = node->pos[k];
	      }
	      if(next_pos[j] < next_pos[k]){
		diag_next_pos = next_pos[j];
	      }else{
		diag_next_pos = next_pos[k];
	      }
	      if(diag_pos != diag_next_pos){
		if(external_gap_penalty > internal_gap_penalty){
		  est_diff = bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][diag_pos]
		    * alignment[j].dim
		      + bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][diag_pos]
			* alignment[k].dim;
		}else{
		  est_diff = bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][diag_pos]
		    * alignment[j].dim
		      + bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][diag_pos]
			* alignment[k].dim;
		}
		n = 0;
		for(l = 0; l < alignment[j].dim; l++){
		  n += bmin_cost[alignment[j].seq[l].str[diag_pos]
				 - 'A'][k][diag_pos];
		}
		if(n < est_diff){
		  est_diff = n;
		}
		n = 0;
		for(m = 0; m < alignment[k].dim; m++){
		  n += bmin_cost[alignment[k].seq[m].str[diag_pos]
				 - 'A'][j][diag_pos];
		}
		if(n < est_diff){
		  est_diff = n;
		}
		edge_cost -= bs_k * est_diff;
	      }
	      if(external_gap_penalty > internal_gap_penalty){
		if(node->pos[j] < node->pos[k]){
		  edge_cost
		    += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		      * (node->pos[k] - node->pos[j]);
		}else{
		  edge_cost
		    += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][0]
		      * (node->pos[j] - node->pos[k]);
		}
		if(next_pos[j] < next_pos[k]){
		  edge_cost
		    -= bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		      * (next_pos[k] - next_pos[j]);
		}else{
		  edge_cost
		    -= bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][0]
		      * (next_pos[j] - next_pos[k]);
		}
	      }else{
		if(node->pos[j] < node->pos[k]){
		  edge_cost
		    += bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][0]
		      * (node->pos[k] - node->pos[j]);
		}else{
		  edge_cost
		    += bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][0]
		      * (node->pos[j] - node->pos[k]);
		}
		if(next_pos[j] < next_pos[k]){
		  edge_cost
		    -= bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][0]
		      * (next_pos[k] - next_pos[j]);
		}else{
		  edge_cost
		    -= bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][0]
		      * (next_pos[j] - next_pos[k]);
		}
	      }
	      edge_cost -= bs_k * add_cost * ((next_pos[j] - node->pos[j])
					      + (next_pos[k] - node->pos[k]));
	    }
	  }
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(fs_info(next_node) == NULL){
	  fs_info_init(next_node);
	}
	if(is_bs_finished(next_node)){
	  is_bridge = 1;
	}
#ifdef VERBOSE
	if(lookup_fs_status(next_node) == UNVISITED){
	  fs_visited++;
	  if((fs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (forward search)\n", fs_visited);
	  }
	}
#endif
	extend_node(fs_pqueue, node, next_node, edge_cost);
      fs_next_edge:
      }
      if(is_bridge){
	bridge_node_list = node_list_cons(node, bridge_node_list);
      }
    }
    if(dir_mask & BMASK){
      node = get_node_from_pqueue(bs_pqueue);
      set_bs_status(node, FINISHED);
#ifdef VERBOSE
      bs_expanded++;
#endif
      if((node == orig_node) || is_fs_finished(node)){
	mid_node = node;
	break;
      }
      for(i = 1; i < (1 << dim); i++){
	n = 1;
	for(j = 0; j < dim; j++){
	  if(i & n){
	    if((next_pos[j] = node->pos[j] - 1) < 0){
	      goto bs_next_edge;
	    }
	  }else{
	    next_pos[j] = node->pos[j];
	  }
	  n <<= 1;
	}
	edge_cost = 0;
	for(j = 0; j < dim; j++){
	  for(k = j + 1; k < dim; k ++){
	    if((node->pos[j] != next_pos[j]) && (node->pos[k] != next_pos[k])){
	      for(l = 0; l < alignment[j].dim; l++){
		for(m = 0; m < alignment[k].dim; m++){
		  edge_cost += wk
		    * matching_cost(alignment[j].seq[l].str[next_pos[j]],
				    alignment[k].seq[m].str[next_pos[k]]);
		}
	      }
	      edge_cost += wk * add_cost * 2;
	    }else if(node->pos[j] != next_pos[j]){
	      n = 0; 
	      for(l = 0; l < alignment[j].dim; l++){
		if(is_normal_char(alignment[j].seq[l].str[next_pos[j]])){
		  n++;
		}
	      }
	      for(m = 0; m < alignment[k].dim; m++){
		if(((next_pos[k] == 0)
		    || (alignment[k].seq[m].str[next_pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[next_pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }else if(node->pos[k] != next_pos[k]){
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		if(is_normal_char(alignment[k].seq[m].str[next_pos[k]])){
		  n++;
		}
	      }
	      for(l = 0; l < alignment[j].dim; l++){
		if(((next_pos[j] == 0)
		    || (alignment[j].seq[l].str[next_pos[j] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((next_pos[j] == alignment[j].seq[0].len)
		       || (alignment[j].seq[l].str[next_pos[j]]
			   == EXTERNAL_GAP_CHAR))){
		  edge_cost += wk * n * (-external_gap_penalty);
		}else{
		  edge_cost += wk * n * (-internal_gap_penalty);
		}
	      }
	      edge_cost += wk * add_cost;
	    }
	    if(dir_mask & FMASK){
	      if(alignment[j].seq[0].len - node->pos[j]
		 < alignment[k].seq[0].len - node->pos[k]){
		diag_pos = alignment[j].seq[0].len - node->pos[j];
	      }else{
		diag_pos = alignment[k].seq[0].len - node->pos[k];
	      }
	      if(alignment[j].seq[0].len - next_pos[j]
		 < alignment[k].seq[0].len - next_pos[k]){
		diag_next_pos = alignment[j].seq[0].len - next_pos[j];
	      }else{
		diag_next_pos = alignment[k].seq[0].len - next_pos[k];
	      }
	      if(diag_pos != diag_next_pos){
		if(external_gap_penalty > internal_gap_penalty){
		  est_diff = fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - diag_next_pos]
		      * alignment[j].dim
			+ fmin_cost[EXTERNAL_GAP_CHAR - 'A']
			  [j][alignment[j].seq[0].len - diag_next_pos]
			    * alignment[k].dim;
		}else{
		  est_diff = fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [k][alignment[k].seq[0].len - diag_next_pos]
		      * alignment[j].dim
			+ fmin_cost[INTERNAL_GAP_CHAR - 'A']
			  [j][alignment[j].seq[0].len - diag_next_pos]
			    * alignment[k].dim;
		}
		n = 0;
		for(l = 0; l < alignment[j].dim; l++){
		  n += fmin_cost[alignment[j].seq[l]
				 .str[alignment[j].seq[0].len
				      - diag_next_pos] - 'A']
					[k][alignment[k].seq[0].len
					    - diag_next_pos];
		}
		if(n < est_diff){
		  est_diff = n;
		}
		n = 0;
		for(m = 0; m < alignment[k].dim; m++){
		  n += fmin_cost[alignment[k].seq[m]
				 .str[alignment[k].seq[0].len
				      - diag_next_pos] - 'A']
					[j][alignment[j].seq[0].len
					    - diag_next_pos];
		}
		if(n < est_diff){
		  est_diff = n;
		}
		edge_cost -= fs_k * est_diff;
	      }
	      if(external_gap_penalty > internal_gap_penalty){
		if(alignment[j].seq[0].len - next_pos[j]
		   < alignment[k].seq[0].len - next_pos[k]){
		  edge_cost
		    -= fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [k][alignment[k].seq[0].len - 1]
			* ((alignment[k].seq[0].len - next_pos[k])
			   - (alignment[j].seq[0].len - next_pos[j]));
		}else{
		  edge_cost
		    -= fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - 1]
			* ((alignment[j].seq[0].len - next_pos[j])
			   - (alignment[k].seq[0].len - next_pos[k]));
		}
		if(alignment[j].seq[0].len - node->pos[j]
		   < alignment[k].seq[0].len - node->pos[k]){
		  edge_cost
		    += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [k][alignment[k].seq[0].len - 1]
			* ((alignment[k].seq[0].len - node->pos[k])
			   - (alignment[j].seq[0].len - node->pos[j]));
		}else{
		  edge_cost
		    += fs_k * fmin_cost[EXTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - 1]
			* ((alignment[j].seq[0].len - node->pos[j])
			   - (alignment[k].seq[0].len - node->pos[k]));
		}
	      }else{
		if(alignment[j].seq[0].len - next_pos[j]
		   < alignment[k].seq[0].len - next_pos[k]){
		  edge_cost
		    -= fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		      [k][alignment[k].seq[0].len - 1]
			* ((alignment[k].seq[0].len - next_pos[k])
			   - (alignment[j].seq[0].len - next_pos[j]));
		}else{
		  edge_cost
		  -= fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		    [j][alignment[j].seq[0].len - 1]
		      * ((alignment[j].seq[0].len - next_pos[j])
			 - (alignment[k].seq[0].len - next_pos[k]));
		}
		if(alignment[j].seq[0].len - node->pos[j]
		   < alignment[k].seq[0].len - node->pos[k]){
		  edge_cost
		    += fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		      [k][alignment[k].seq[0].len - 1]
			* ((alignment[k].seq[0].len - node->pos[k])
			   - (alignment[j].seq[0].len - node->pos[j]));
		}else{
		  edge_cost
		    += fs_k * fmin_cost[INTERNAL_GAP_CHAR - 'A']
		      [j][alignment[j].seq[0].len - 1]
			* ((alignment[j].seq[0].len - node->pos[j])
			   - (alignment[k].seq[0].len - node->pos[k]));
		}
	      }
	      edge_cost -= fs_k * add_cost * ((node->pos[j] - next_pos[j])
					      + (node->pos[k] - next_pos[k]));
	    }
	    if(node->pos[j] < node->pos[k]){
	      diag_pos = node->pos[j];
	    }else{
	      diag_pos = node->pos[k];
	    }
	    if(next_pos[j] < next_pos[k]){
	      diag_next_pos = next_pos[j];
	    }else{
	      diag_next_pos = next_pos[k];
	    }
	    if(diag_pos != diag_next_pos){
	      if(external_gap_penalty > internal_gap_penalty){
		est_diff = bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][diag_next_pos]
		  * alignment[j].dim
		    + bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][diag_next_pos]
		      * alignment[k].dim;
	      }else{
		est_diff = bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][diag_next_pos]
		  * alignment[j].dim
		    + bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][diag_next_pos]
		      * alignment[k].dim;
	      }
	      n = 0;
	      for(l = 0; l < alignment[j].dim; l++){
		n += bmin_cost[alignment[j].seq[l].str[diag_next_pos]
			       - 'A'][k][diag_next_pos];
	      }
	      if(n < est_diff){
		est_diff = n;
	      }
	      n = 0;
	      for(m = 0; m < alignment[k].dim; m++){
		n += bmin_cost[alignment[k].seq[m].str[diag_next_pos]
			       - 'A'][j][diag_next_pos];
	      }
	      if(n < est_diff){
		est_diff = n;
	      }
	      edge_cost -= bs_k * est_diff;
	    }
	    if(external_gap_penalty > internal_gap_penalty){
	      if(next_pos[j] < next_pos[k]){
		edge_cost
		  += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		    * (next_pos[k] - next_pos[j]);
	      }else{
		edge_cost
		  += bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][0]
		    * (next_pos[j] - next_pos[k]);
	      }
	      if(node->pos[j] < node->pos[k]){
		edge_cost
		  -= bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][j][0]
		    * (node->pos[k] - node->pos[j]);
	      }else{
		edge_cost
		  -= bs_k * bmin_cost[EXTERNAL_GAP_CHAR - 'A'][k][0]
		    * (node->pos[j] - node->pos[k]);
	      }
	    }else{
	      if(next_pos[j] < next_pos[k]){
		edge_cost
		  += bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][0]
		    * (next_pos[k] - next_pos[j]);
	      }else{
		edge_cost
		  += bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][0]
		    * (next_pos[j] - next_pos[k]);
	      }
	      if(node->pos[j] < node->pos[k]){
		edge_cost
		  -= bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][j][0]
		    * (node->pos[k] - node->pos[j]);
	      }else{
		edge_cost
		  -= bs_k * bmin_cost[INTERNAL_GAP_CHAR - 'A'][k][0]
		    * (node->pos[j] - node->pos[k]);
	      }
	    }
	    edge_cost -= bs_k * add_cost * ((node->pos[j] - next_pos[j])
					    + (node->pos[k] - next_pos[k]));
	  }
	}
	if((next_node = get_node(node_set, next_pos)) == NULL){
	  next_node = create_node(node_set, next_pos);
	}
	if(bs_info(next_node) == NULL){
	  bs_info_init(next_node);
	}
	if(is_fs_finished(next_node)){
	  bridge_node_list = node_list_cons(next_node, bridge_node_list);
	}
#ifdef VERBOSE
	if(lookup_bs_status(next_node) == UNVISITED){
	  bs_visited++;
	  if((bs_visited & 0x7fff) == 0){
	    fprintf(stderr,
		    ".... visited %d nodes (backward search)\n", bs_visited);
	  }
       	}
#endif
	extend_node(bs_pqueue, node, next_node, edge_cost);
      bs_next_edge:
      }
    }
  }
#ifdef VERBOSE
  times(&t_search_end);
#endif
  if(dir_mask & FMASK){
    for(i = 'A'; i <= 'Z' + 2; i++){
      for(j = 0; j < dim; j++){
	free(fmin_cost[i - 'A'][j]);
      }
    }
    pqueue_finish(fs_pqueue);
  }
  if(dir_mask & BMASK){
    for(i = 'A'; i <= 'Z' + 2; i++){
      for(j = 0; j < dim; j++){
	free(bmin_cost[i - 'A'][j]);
      }
    }
    pqueue_finish(bs_pqueue);
  }
  for(node_list = bridge_node_list;
      node_list != NULL; node_list = node_list->next){
    if(lookup_fs_cost(node_list->node) + lookup_bs_cost(node_list->node)
       < lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node)){
      mid_node = node_list->node;
    }
  }
  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  al_p->dim = 0;
  for(i = 0; i < dim; i++){
    al_p->dim += alignment[i].dim;
  }
  al_len = 0;
  for(node = mid_node; node != orig_node; node = lookup_fs_prev_node(node)){
    al_len++;
  }
  for(node = mid_node; node != dest_node; node = lookup_bs_prev_node(node)){
    al_len++;
  }
  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * al_p->dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  k = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      al_p->seq[k].len = al_len;
      if((al_p->seq[k].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
      al_p->seq[k].str[al_len] = '\0';
      p = -1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	p++;
      }
      q = mid_node->pos[i] - 1;
      for(node = mid_node; node != orig_node;
	  node = lookup_fs_prev_node(node)){
	if(node->pos[i] != lookup_fs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	}else{
	  if(((q + 1 > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q + 1] == EXTERNAL_GAP_CHAR))
	     || ((q < 0)
		 || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p--] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p--] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      p = al_len;
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	p--;
      }
      q = mid_node->pos[i];
      for(node = mid_node; node != dest_node;
	  node = lookup_bs_prev_node(node)){
	if(node->pos[i] != lookup_bs_prev_node(node)->pos[i]){
	  al_p->seq[k].str[p++] = alignment[i].seq[j].str[q++];
	}else{
	  if(((q > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))
	     || ((q - 1 < 0)
		 || (alignment[i].seq[j].str[q - 1] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p++] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p++] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      k++;
    }
  }
  al_score = -((lookup_fs_cost(mid_node) + lookup_bs_cost(mid_node)) / wk);
#ifdef VERBOSE
  fprintf(stderr, "Estimated inter-group cost: %d\n",
	  (lookup_fs_cost(orig_node) + lookup_bs_cost(dest_node)) / wk);
  fprintf(stderr, "Actual inter-group cost: %d\n", -al_score);
#endif
#ifndef NO_ITER
  node_set_finish(node_set);
#endif
  for(i = 0; i < dim; i++){
    for(j = i + 1; j < dim; j++){
      al_score
	+= add_cost * (alignment[i].seq[0].len + alignment[j].seq[0].len);
    }
  }
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      for(k = j + 1; k < alignment[i].dim; k++){
	for(l = 0; l < alignment[i].seq[0].len; l++){
	  al_score += matching_cost(alignment[i].seq[j].str[l],
				    alignment[i].seq[k].str[l]);
	}
      }
    }
  }
  al_p->score = al_score;
#ifdef VERBOSE
  times(&t_end);
  wnodes = alignment[0].seq[0].len + 1;
  for(i = 1; i < dim; i++){
    wnodes *= alignment[i].seq[0].len + 1;
  }
  fprintf(stderr, "# of whole nodes: %8.2e\n", wnodes);
  fprintf(stderr, "# of visited nodes (forward search): %d\n", fs_visited);
  fprintf(stderr, "# of expanded nodes (forward search): %d\n", fs_expanded);
  fprintf(stderr, "# of visited nodes (backward search): %d\n", bs_visited);
  fprintf(stderr, "# of expanded nodes (backward search): %d\n", bs_expanded);
  fprintf(stderr, "Used time:     User    System  Total\n");
#if defined(linux)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / CLK_TCK,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / CLK_TCK,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / CLK_TCK);
#elif defined(sun)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / HZ);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / HZ,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / HZ,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime)) / HZ);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / HZ);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / HZ);
#else
  fprintf(stderr, "  Pre-process  %6d  %6d  %6d\n",
	  (int)(t_search_start.tms_utime - t_start.tms_utime),
	  (int)(t_search_start.tms_stime - t_start.tms_stime),
	  (int)((t_search_start.tms_utime - t_start.tms_utime)
		+ (t_search_start.tms_stime - t_start.tms_stime)));
  fprintf(stderr, "  Search       %6d  %6d  %6d\n",
	  (int)(t_search_end.tms_utime - t_search_start.tms_utime),
	  (int)(t_search_end.tms_stime - t_search_start.tms_stime),
	  (int)((t_search_end.tms_utime - t_search_start.tms_utime)
		+ (t_search_end.tms_stime - t_search_start.tms_stime)));
  fprintf(stderr, "  Post-process %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_search_end.tms_utime),
	  (int)(t_end.tms_stime - t_search_end.tms_stime),
	  (int)((t_end.tms_utime - t_search_end.tms_utime)
		+ (t_end.tms_stime - t_search_end.tms_stime)));
  fprintf(stderr, "  Total        %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_start.tms_utime),
	  (int)(t_end.tms_stime - t_start.tms_stime),
	  (int)((t_end.tms_utime - t_start.tms_utime)
		+ (t_end.tms_stime - t_start.tms_stime)));
#endif
#endif
  return al_p;
}
