#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

#define h_e_parent(n) ((n) / 2)
#define h_e_left(n) ((n) * 2)
#define h_e_right(n) ((n) * 2 + 1)

#define DEBUG_1 0

void make_histogram_for_h_t(epp_node_t eppnode, int parent_delta, int *histogram,
     node_t mid_node, node_t orig_node, node_t dest_node,
     node_set_t node_set, int dim, alignment_t *alignment, 
     int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
     int *constructed, int *h_t, int *h_edge,
#endif
     int add_cost, estimator_t *estimator,
     int print_subopt, subopt_answer_t subanswer, int depth,
     int *maxdepth_p, int line, solutions_p opt_sols,
     solutions2_p opt_sols2, solutions2_p opt_sols3)
{
  epp_node_t ht;
  int tmp_delta;
  node_t tmp_node;
  int tmp;

  if (eppnode == NULL) return;
  if ((tmp_delta = parent_delta + lookup_ht_delta(eppnode, dir)) <= delta * wk) {
#ifdef DEBUG_1
    if (tmp_delta < 0) {
      fprintf(stderr," Error detected : delta\n");
    } else {
#endif
      (histogram[(tmp_delta / wk)])++;
#ifdef DEBUG_1
    }
#endif
  } else return;

  construct_heap((tmp_node = lookup_ht_node(eppnode, dir)),
		 node_set, dim, alignment, 
		 wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		 constructed, h_t, h_edge,
#endif
		 add_cost, estimator, CONSTRUCTED);

  if (print_subopt || (opt_sols!=NULL) ||
      (opt_sols2!=NULL) || (opt_sols3!=NULL)) {
    if ((*maxdepth_p) <= depth) {
      (*maxdepth_p) *= 2;
      if ((realloc_subopt_answer(subanswer, (*maxdepth_p)))==NULL) {
	fprintf(stderr, "\nCannot realloc for suboptimal asnwers(%d : %d -> %d)\n",
		depth, (*maxdepth_p)/2, (*maxdepth_p));
	exit(1);
      }
    }
    set_f_of_subopt_answer(subanswer, depth, eppnode->node);
    set_t_of_subopt_answer(subanswer, depth, tmp_node);
    if (print_subopt) printf("Delta: %d\n",(tmp_delta / wk));
    print_subopt_answer(subanswer, depth, mid_node, orig_node, dest_node, dir,
			dim, alignment, line, print_subopt,
			opt_sols, opt_sols2, opt_sols3);
  }

  if ((ht = lookup_tree_heap(tmp_node, dir)) != NULL) {
    make_histogram_for_h_t(ht, tmp_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth+1, maxdepth_p, line,
			   opt_sols, opt_sols2, opt_sols3);
  }

  if ((ht = eppnode->left) != NULL) {
    make_histogram_for_h_t(ht, parent_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth, maxdepth_p, line,
			   opt_sols, opt_sols2, opt_sols3);
  }

  if ((ht = eppnode->right) != NULL) {
    make_histogram_for_h_t(ht, parent_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth, maxdepth_p, line,
			   opt_sols, opt_sols2, opt_sols3);
  }

  if ((tmp = lookup_edge_heap_size(eppnode->node, dir)) > 1) {
    make_histogram_for_h_edge(lookup_edge_heap(eppnode->node, dir), eppnode->node,
			      tmp, 1, parent_delta, histogram,
			      mid_node, orig_node, dest_node,
			      node_set, dim, alignment, 
			      wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			      constructed, h_t, h_edge,
#endif
			      add_cost, estimator,
			      print_subopt, subanswer, depth, maxdepth_p, line,
			      opt_sols, opt_sols2, opt_sols3);
  }
}


void make_histogram_for_h_t_d1(epp_node_t eppnode, int parent_delta, int *histogram,
     node_t mid_node, node_t orig_node, node_t dest_node,
     node_set_t node_set, int dim, alignment_t *alignment, 
     int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
     int *constructed, int *h_t, int *h_edge,
#endif
     int add_cost, estimator_t *estimator,
     int print_subopt, subopt_answer_t subanswer, int depth, int *maxdepth_p, int line)
{
  epp_node_t ht;
  int tmp_delta;
  node_t tmp_node;
  int tmp;

  if (eppnode == NULL) return;
  if ((tmp_delta = parent_delta + lookup_ht_delta(eppnode, dir)) <= delta * wk) {
#ifdef DEBUG_1
    if (tmp_delta < 0) {
      fprintf(stderr," Error detected : delta\n");
    } else {
#endif
      (histogram[(tmp_delta / wk)])++;
#ifdef DEBUG_1
    }
#endif
  } else return;

  construct_heap((tmp_node = lookup_ht_node(eppnode, dir)), node_set, dim, alignment, 
		 wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		 constructed, h_t, h_edge,
#endif
		 add_cost, estimator, CONSTRUCTED);

  if (print_subopt) {
    if ((*maxdepth_p) <= depth) {
      (*maxdepth_p) *= 2;
      if ((realloc_subopt_answer(subanswer, (*maxdepth_p)))==NULL) {
	fprintf(stderr, "\nCannot realloc for suboptimal asnwers(%d : %d -> %d)\n", depth, (*maxdepth_p)/2, (*maxdepth_p));
	exit(1);
      }
    }
    set_f_of_subopt_answer(subanswer, depth, eppnode->node);
    set_t_of_subopt_answer(subanswer, depth, tmp_node);
    if (print_subopt) printf("Delta: %d\n",(tmp_delta / wk));
    print_subopt_answer(subanswer, depth, mid_node, orig_node, dest_node, dir,
			dim, alignment, line, 1, NULL, NULL, NULL);
  }

  if ((ht = lookup_tree_heap(tmp_node, dir)) != NULL) {
    make_histogram_for_h_t_d2(ht, tmp_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth+1, maxdepth_p, line);
  }

  if ((ht = eppnode->left) != NULL) {
    make_histogram_for_h_t_d1(ht, parent_delta, histogram,
			      mid_node, orig_node, dest_node,
			      node_set, dim, alignment, 
			      wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			      constructed, h_t, h_edge,
#endif
			      add_cost, estimator,
			      print_subopt, subanswer, depth, maxdepth_p, line);
  }

  if ((ht = eppnode->right) != NULL) {
    make_histogram_for_h_t_d1(ht, parent_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth, maxdepth_p, line);
  }

  if ((tmp = lookup_edge_heap_size(eppnode->node, dir)) > 1) {
    make_histogram_for_h_edge_d(lookup_edge_heap(eppnode->node, dir), eppnode->node,
				tmp, 1, parent_delta, histogram,
				mid_node, orig_node, dest_node,
				node_set, dim, alignment, 
				wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
				constructed, h_t, h_edge,
#endif
				add_cost, estimator,
				print_subopt, subanswer, depth, maxdepth_p, line);
  }
}


void make_histogram_for_h_t_d2(epp_node_t eppnode, int parent_delta, int *histogram,
     node_t mid_node, node_t orig_node, node_t dest_node,
     node_set_t node_set, int dim, alignment_t *alignment, 
     int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
     int *constructed, int *h_t, int *h_edge,
#endif
     int add_cost, estimator_t *estimator,
     int print_subopt, subopt_answer_t subanswer, int depth, int *maxdepth_p, int line)
{
  epp_node_t ht;
  int tmp_delta;
  node_t tmp_node;
  int tmp;
  int not_on_the_best_path;

  if (eppnode == NULL) return;
  if ((tmp_delta = parent_delta + lookup_ht_delta(eppnode, dir)) <= delta * wk) {
#ifdef DEBUG_1
    if (tmp_delta < 0) {
      fprintf(stderr," Error detected : delta\n");
    } else {
#endif
      if (not_on_the_best_path = (!(is_on_the_best_path(eppnode->node, dir))))
	(histogram[(tmp_delta / wk)])++;
#ifdef DEBUG_1
    }
#endif
  } else return;

  if (not_on_the_best_path) {
    construct_heap((tmp_node = lookup_ht_node(eppnode, dir)), node_set, dim, alignment, 
		   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		   constructed, h_t, h_edge,
#endif
		   add_cost, estimator, CONSTRUCTED);

    if (print_subopt) {
      if ((*maxdepth_p) <= depth) {
	(*maxdepth_p) *= 2;
	if ((realloc_subopt_answer(subanswer, (*maxdepth_p)))==NULL) {
	  fprintf(stderr, "\nCannot realloc for suboptimal asnwers(%d : %d -> %d)\n", depth, (*maxdepth_p)/2, (*maxdepth_p));
	  exit(1);
	}
      }
      set_f_of_subopt_answer(subanswer, depth, eppnode->node);
      set_t_of_subopt_answer(subanswer, depth, tmp_node);
      if (print_subopt) printf("Delta: %d\n",(tmp_delta / wk));
      print_subopt_answer(subanswer, depth, mid_node, orig_node, dest_node, dir,
			  dim, alignment, line, 1, NULL, NULL, NULL);
    }

    if ((ht = lookup_tree_heap(tmp_node, dir)) != NULL) {
      make_histogram_for_h_t_d2(ht, tmp_delta, histogram,
				mid_node, orig_node, dest_node,
				node_set, dim, alignment, 
				wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
				constructed, h_t, h_edge,
#endif
				add_cost, estimator,
				print_subopt, subanswer, depth+1, maxdepth_p, line);
    }

    if ((tmp = lookup_edge_heap_size(eppnode->node, dir)) > 1) {
      make_histogram_for_h_edge_d(lookup_edge_heap(eppnode->node, dir), eppnode->node,
				  tmp, 1, parent_delta, histogram,
				  mid_node, orig_node, dest_node,
				  node_set, dim, alignment, 
				  wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
				  constructed, h_t, h_edge,
#endif
				  add_cost, estimator,
				  print_subopt, subanswer, depth, maxdepth_p, line);
    }
  }

  if ((ht = eppnode->left) != NULL) {
    make_histogram_for_h_t_d2(ht, parent_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth, maxdepth_p, line);
  }

  if ((ht = eppnode->right) != NULL) {
    make_histogram_for_h_t_d2(ht, parent_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth, maxdepth_p, line);
  }

}


void make_histogram_for_h_edge(edge_heap_t edgeheap, node_t parent_node,
     int size, int number, int parent_delta, int *histogram,
     node_t mid_node, node_t orig_node, node_t dest_node,
     node_set_t node_set, int dim, alignment_t *alignment, 
     int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
     int *constructed, int *h_t, int *h_edge,
#endif
     int add_cost, estimator_t *estimator,
     int print_subopt, subopt_answer_t subanswer, int depth,
     int *maxdepth_p, int line, solutions_p opt_sols,
     solutions2_p opt_sols2, solutions2_p opt_sols3)
{
  epp_node_t ht;
  int tmp_delta;
  node_t tmp_node;
  int child;

  if ((tmp_delta = parent_delta + edgeheap[number].delta) <= delta * wk) {
    (histogram[(tmp_delta / wk)])++;
  } else return;

  construct_heap((tmp_node = edgeheap[number].node), node_set, dim, alignment, 
		 wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		 constructed, h_t, h_edge,
#endif
		 add_cost, estimator, CONSTRUCTED);

  if (print_subopt || (opt_sols!=NULL) || (opt_sols2!=NULL) || (opt_sols3!=NULL)) {
    if ((*maxdepth_p) <= depth) {
      (*maxdepth_p) *= 2;
      if ((realloc_subopt_answer(subanswer, (*maxdepth_p)))==NULL) {
	fprintf(stderr, "\nCannot realloc for suboptimal asnwers(%d : %d -> %d)\n",
		depth, (*maxdepth_p)/2, (*maxdepth_p));
	exit(1);
      }
    }
    set_f_of_subopt_answer(subanswer, depth, parent_node);
    set_t_of_subopt_answer(subanswer, depth, tmp_node);
    if (print_subopt) printf("Delta: %d\n",(tmp_delta / wk));
    print_subopt_answer(subanswer, depth, mid_node, orig_node, dest_node, dir,
			dim, alignment, line, print_subopt,
			opt_sols, opt_sols2, opt_sols3);
  }

  if ((ht = lookup_tree_heap(tmp_node, dir)) != NULL) {
    make_histogram_for_h_t(ht, tmp_delta, histogram,
			   mid_node, orig_node, dest_node,
                           node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth+1, maxdepth_p, line,
			   opt_sols, opt_sols2, opt_sols3);
  }

  if ((child = h_e_left(number)) < size) {
    make_histogram_for_h_edge(edgeheap, parent_node, size, child,
                              parent_delta, histogram,
			      mid_node, orig_node, dest_node,
			      node_set, dim, alignment, 
			      wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			      constructed, h_t, h_edge,
#endif
			      add_cost, estimator,
			      print_subopt, subanswer, depth, maxdepth_p, line,
			      opt_sols, opt_sols2, opt_sols3);
  }

  child++;
  if (child < size) {
    make_histogram_for_h_edge(edgeheap, parent_node, size, child,
			      parent_delta, histogram,
			      mid_node, orig_node, dest_node,
			      node_set, dim, alignment, 
			      wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			      constructed, h_t, h_edge,
#endif
			      add_cost, estimator,
			      print_subopt, subanswer, depth, maxdepth_p, line,
			      opt_sols, opt_sols2, opt_sols3);
  }
}

void make_histogram_for_h_edge_d(edge_heap_t edgeheap, node_t parent_node,
     int size, int number, int parent_delta, int *histogram,
     node_t mid_node, node_t orig_node, node_t dest_node,
     node_set_t node_set, int dim, alignment_t *alignment, 
     int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
     int *constructed, int *h_t, int *h_edge,
#endif
     int add_cost, estimator_t *estimator,
     int print_subopt, subopt_answer_t subanswer, int depth, int *maxdepth_p, int line)
{
  epp_node_t ht;
  int tmp_delta;
  node_t tmp_node;
  int child;

  if ((tmp_delta = parent_delta + edgeheap[number].delta) <= delta * wk) {
    (histogram[(tmp_delta / wk)])++;
  } else return;

  construct_heap((tmp_node = edgeheap[number].node), node_set, dim, alignment, 
		 wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		 constructed, h_t, h_edge,
#endif
		 add_cost, estimator, CONSTRUCTED);

  if (print_subopt) {
    if ((*maxdepth_p) <= depth) {
      (*maxdepth_p) *= 2;
      if ((realloc_subopt_answer(subanswer, (*maxdepth_p)))==NULL) {
	fprintf(stderr, "\nCannot realloc for suboptimal asnwers(%d : %d -> %d)\n", depth, (*maxdepth_p)/2, (*maxdepth_p));
	exit(1);
      }
    }
    set_f_of_subopt_answer(subanswer, depth, parent_node);
    set_t_of_subopt_answer(subanswer, depth, tmp_node);
    if (print_subopt) printf("Delta: %d\n",(tmp_delta / wk));
    print_subopt_answer(subanswer, depth, mid_node, orig_node, dest_node, dir,
			dim, alignment, line, 1, NULL, NULL, NULL);
  }

  if ((ht = lookup_tree_heap(tmp_node, dir)) != NULL) {
    make_histogram_for_h_t_d2(ht, tmp_delta, histogram,
			   mid_node, orig_node, dest_node,
			   node_set, dim, alignment, 
			   wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
			   constructed, h_t, h_edge,
#endif
			   add_cost, estimator,
			   print_subopt, subanswer, depth+1, maxdepth_p, line);
  }

  if ((child = h_e_left(number)) < size) {
    make_histogram_for_h_edge_d(edgeheap, parent_node, size, child,
				parent_delta, histogram,
				mid_node, orig_node, dest_node,
				node_set, dim, alignment, 
				wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
				constructed, h_t, h_edge,
#endif
				add_cost, estimator,
				print_subopt, subanswer, depth, maxdepth_p, line);
  }

  child++;
  if (child < size) {
    make_histogram_for_h_edge_d(edgeheap, parent_node, size, child,
				parent_delta, histogram,
				mid_node, orig_node, dest_node,
				node_set, dim, alignment, 
				wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
				constructed, h_t, h_edge,
#endif
				add_cost, estimator,
				print_subopt, subanswer, depth, maxdepth_p, line);
  }
}
