#include <stdio.h>
#include <stdlib.h>
#include "align.h"

extern node_t node_init(int, int *);
extern void node_finish(node_t);

#ifndef DEFAULT_TABLE_SIZE
#ifdef SAVE_MEMORY
#define DEFAULT_TABLE_SIZE 100003
#else
#define DEFAULT_TABLE_SIZE 2000003
/*
#define DEFAULT_TABLE_SIZE 500009
#define DEFAULT_TABLE_SIZE 3999971
*/
#endif
#endif

#define BASE 787

node_set_t node_set_init(int dim)
{
  node_set_t t;

  if((t = (node_set_t)malloc(sizeof(struct node_set))) == NULL){
    fprintf(stderr, "Can't allocate memory for node-set\n");
    exit(-1);
  }
  t->dim = dim;
  t->count = 0;
  t->table_size = DEFAULT_TABLE_SIZE;
  if((t->node_list_table
      = (node_list_t *)calloc(t->table_size, sizeof(node_list_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for node-set\n");
    exit(-1);
  }
#ifdef DEBUG
  t->long_list_len = 0;
#endif
  return t;
}

void node_set_finish(node_set_t node_set)
{
#ifndef OMIT_FREE
  int i;
  node_list_t t;

  for(i = 0; i < node_set->table_size; i++){
    for(t = node_set->node_list_table[i]; t != NULL; t = node_list_decons(t)){
      node_finish(t->node);
    }
  }
#endif
  free(node_set->node_list_table);
  free(node_set);
}

void node_set_reset(node_set_t node_set)
{
  int i;
  node_list_t t;

  for(i = 0; i < node_set->table_size; i++){
    for(t = node_set->node_list_table[i]; t != NULL; t = t->next){
      if(fs_info(t->node) != NULL){
	fs_info_reset(t->node);
      }
      if(bs_info(t->node) != NULL){
	bs_info_reset(t->node);
      }
    }
  }
}

unsigned int hash(node_set_t node_set, int *pos)
{
  int i;
  unsigned int t;

  t = pos[0] % node_set->table_size;
  for(i = 1; i < node_set->dim; i++){
    t = (t * BASE + pos[i]) % node_set->table_size;
  }
  return t;
}

node_t get_node(node_set_t node_set, int *pos)
{
  int i;
#ifdef DEBUG
  int n;
#endif
  node_list_t t;

#ifdef DEBUG
  n = 1;
#endif
  for(t = node_set->node_list_table[hash(node_set, pos)];
      t != NULL; t = t->next){
    i = 0;
    while(t->node->pos[i] == pos[i]){
      if(++i >= node_set->dim){
#ifdef DEBUG
	if(n > node_set->long_list_len){
	  node_set->long_list_len = n;
	  fprintf(stderr, "Long node-list found: %d\n", n);
	}
#endif
	return t->node;
      }
    }
#ifdef DEBUG
    n++;
#endif
  }
  return NULL;
}

node_t create_node(node_set_t node_set, int *pos)
{
  int i;
  node_t t;

  t = node_init(node_set->dim, pos);
  i = hash(node_set, pos);
  node_set->node_list_table[i]
    = node_list_cons(t, node_set->node_list_table[i]);
  node_set->count++;
  return t;
}
