#include <stdio.h>
#include <stdlib.h>
#include "align.h"

void iter_parametric_gap(int dim, alignment_t *alignment,
   int wk, int fs_k, int bs_k, int dir_mask, int line,
   int print_subopt, solutions_p boundaries, int left, int right,
   int *max_iteration_p, int *next_iteration_p)
{
  int left_score, right_score;
  int left_gap_penalty, right_gap_penalty;
  int left_gap_denominator, right_gap_denominator;
  int left_gap, right_gap;
  int left_base, right_base;
  int intersect_gap_penalty;
  int new_gap_penalty, new_gap_denominator;
  int new_solutions;
  int g;
  int new_lbound, lb1, lb2;

  left_gap = boundaries[left].max_gap;
  right_gap = boundaries[right].min_gap;

  if (left_gap == right_gap) return;

  left_score = boundaries[left].score;
  right_score = boundaries[right].score;

  left_gap_penalty = boundaries[left].gap_penalty;
  right_gap_penalty = boundaries[right].gap_penalty;

  left_gap_denominator = boundaries[left].gap_denominator;
  right_gap_denominator = boundaries[right].gap_denominator;

  left_base = (left_score - left_gap_penalty * left_gap) / left_gap_denominator;
  right_base = (right_score - right_gap_penalty * right_gap) / right_gap_denominator;

  new_gap_penalty = left_base - right_base;
  new_gap_denominator = right_gap - left_gap;

  if (new_gap_denominator < 0) {
    new_gap_penalty *= -1;
    new_gap_denominator *= -1;
  }

  g = gcd(new_gap_penalty, new_gap_denominator);

  new_gap_penalty /= g;
  new_gap_denominator /= g;

  internal_gap_penalty = new_gap_penalty;
  external_gap_penalty = new_gap_penalty;
  gap_denominator = new_gap_denominator;

  lb1 = weighted_score_gd(boundaries[left].max_alignment, weights);
  lb2 = weighted_score_gd(boundaries[left].min_alignment, weights);
  if (lb1>lb2) new_lbound = lb1;
  else new_lbound = lb2;

  if ((*max_iteration_p) <= ((*next_iteration_p) + 2)) {
    (*max_iteration_p) *= 2;
    if ((boundaries = (solutions_p)realloc(boundaries, (*max_iteration_p)*sizeof(solutions_t))) == NULL) {
      fprintf(stderr, "Can't re-allocate memory for parametric solutions.\n");
      exit(1);
    }
  }

  pairsubopt(dim, alignment, 1, -new_lbound, wk, fs_k, bs_k, dir_mask, 0, -1,
	     line, 0, print_subopt, (&(boundaries[(*next_iteration_p)])),
	     NULL, NULL);
  new_solutions = (*next_iteration_p);
  boundaries[new_solutions].left = left;
  boundaries[new_solutions].right = right;
  (*next_iteration_p)++;

  iter_parametric_gap(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, left, new_solutions,
              max_iteration_p, next_iteration_p);

  iter_parametric_gap(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, new_solutions, right,
              max_iteration_p, next_iteration_p);
}

void iter_parametric_weight(int dim, alignment_t *alignment,
   int wk, int fs_k, int bs_k, int dir_mask, int line,
   int print_subopt, solutions2_p boundaries, int left, int right,
   int *max_iteration_p, int *next_iteration_p)
{
  int left_slope, right_slope;
  int p, q, d;
  int l1, l2, r1, r2;
  int i, j;
  int new_solutions;
  int new_lbound;

  left_slope = boundaries[left].max_slope;
  right_slope = boundaries[right].min_slope;

  if (left_slope == right_slope) return;

  l1 = weighted_score(boundaries[left].max_alignment, weights1);
  l2 = weighted_score(boundaries[left].max_alignment, weights2);
  r1 = weighted_score(boundaries[right].min_alignment, weights1);
  r2 = weighted_score(boundaries[right].min_alignment, weights2);

  p = l1 - r1;
  q = r2 - l2;
  d = gcd(p, q);
  p /= d;
  q /= d;

  for(i=0;i<dim;i++)
    for(j=0;j<dim;j++)
      weights[i][j] = q * weights1[i][j] + p * weights2[i][j];

  normalize_weights(weights, dim);

  if ((*max_iteration_p) <= ((*next_iteration_p) + 2)) {
    (*max_iteration_p) *= 2;
    if ((boundaries = (solutions2_p)realloc(boundaries, (*max_iteration_p)*sizeof(solutions2_t))) == NULL) {
      fprintf(stderr, "Can't re-allocate memory for parametric solutions.\n");
      exit(1);
    }
  }
  new_lbound = weighted_score(boundaries[left].max_alignment, weights);

  pairsubopt(dim, alignment, 1, -new_lbound, wk, fs_k, bs_k, dir_mask, 0, -1,
	     line, 0, print_subopt, NULL,
	     (&(boundaries[(*next_iteration_p)])), NULL);
  new_solutions = (*next_iteration_p);
  boundaries[new_solutions].p = p;
  boundaries[new_solutions].q = q;
  boundaries[new_solutions].left = left;
  boundaries[new_solutions].right = right;
  (*next_iteration_p)++;

  iter_parametric_weight(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, left, new_solutions,
              max_iteration_p, next_iteration_p);

  iter_parametric_weight(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, new_solutions, right,
              max_iteration_p, next_iteration_p);
}

void iter_parametric_score(int dim, alignment_t *alignment,
   int wk, int fs_k, int bs_k, int dir_mask, int line,
   int print_subopt, solutions2_p boundaries, int left, int right,
   int *max_iteration_p, int *next_iteration_p)
{
  int left_slope, right_slope;
  int p, q, d;
  int l1, l2, r1, r2;
  int i, j;
  int new_solutions;
  int new_lbound;

  left_slope = boundaries[left].max_slope;
  right_slope = boundaries[right].min_slope;

  if (left_slope == right_slope) return;

  l1 = another_score1(boundaries[left].max_alignment);
  l2 = another_score2(boundaries[left].max_alignment);
  r1 = another_score1(boundaries[right].min_alignment);
  r2 = another_score2(boundaries[right].min_alignment);

  p = l1 - r1;
  q = r2 - l2;
  d = gcd(p, q);
  p /= d;
  q /= d;

  for(i=0;i<(('Z'-'A'+1)+2);i++)
    for(j=0;j<(('Z'-'A'+1)+2);j++)
      score_table[i][j] = q * score_table1[i][j] + p * score_table2[i][j];
  max_score = q * max_score1 + p * max_score2;
  internal_gap_penalty = q * internal_gap_penalty1 + p * internal_gap_penalty1;
  external_gap_penalty = q * external_gap_penalty2 + p * external_gap_penalty2;

  printf("--- p = %d, q = %d\n", p, q);

  normalize_score();

  if ((*max_iteration_p) <= ((*next_iteration_p) + 2)) {
    (*max_iteration_p) *= 2;
    if ((boundaries = (solutions2_p)realloc(boundaries, (*max_iteration_p)*sizeof(solutions2_t))) == NULL) {
      fprintf(stderr, "Can't re-allocate memory for parametric solutions.\n");
      exit(1);
    }
  }

  new_lbound = weighted_score(boundaries[left].max_alignment, weights);

  pairsubopt(dim, alignment, 1, -new_lbound, wk, fs_k, bs_k, dir_mask, 0, -1,
	     line, 0, print_subopt, NULL, NULL,
	     (&(boundaries[(*next_iteration_p)])));
  new_solutions = (*next_iteration_p);
  boundaries[new_solutions].p = p;
  boundaries[new_solutions].q = q;
  boundaries[new_solutions].left = left;
  boundaries[new_solutions].right = right;
  (*next_iteration_p)++;

  iter_parametric_score(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, left, new_solutions,
              max_iteration_p, next_iteration_p);

  iter_parametric_score(dim, alignment, wk, fs_k, bs_k, dir_mask, line,
              print_subopt, boundaries, new_solutions, right,
              max_iteration_p, next_iteration_p);
}

int gcd_primitive(int a, int b)   /* must be a >= b >= 0 */
{
  if (b == 0) return (a);
  return gcd_primitive(b, (a % b));
}

int gcd(int a, int b)
{
  if (a < 0) a = -a;
  if (b < 0) b = -b;
  if (a < b) return gcd_primitive(b, a);
  else return gcd_primitive(a, b);
}

int normalize_weights(int **ws, int dim)
{
  int i,j;
  int d;

  d = weights[0][0];
  for(i=0;(i<dim-1);i++) {
    for(j=i+1;j<dim;j++) {
      d = gcd(weights[i][j],d);
      if (d == 1) return 1;
    }
  }
  for(i=0;i<dim;i++)
    for(j=0;j<dim;j++) {
      weights[i][j] /= d;
    }
  return d;
}
